/*
 * PathFinder.java
 *
 * Created on 01 February 2005, 14:46
 */

package MiscTests;

/**
 *
 * @author  Jean-Alain Grunchec
 */
/* Copyright (C) 2002 J. M. Spivey */

import java.awt.Color;
import java.util.*;

/** Object able to carry out a shortest path search using Dijkstra's algorithm.
 *  
 *  Each search is carried out by a freshly-created object. */
class PathFinder {
    private Map map;
        // The map to search in
    private MapViewer viewer;
        // Progress of the search is notified to this MapViewer
    private DistanceWatcher watcher;
        // Another monitor for our progress
    private Town from, to;
        // Starting and finishing towns
    private boolean heur;
        // Whether to adjust distances heuristically to speed the search
    private boolean work;
        // Whether to show working
    private long delay = 0;
        // Delay between steps (for animation)

    public static final float INF = 1.0e6f;
        // That's near enough to infinity
    
    public PathFinder(Map map, Town from, Town to, boolean heur, boolean work) {
        this.map = map;
        this.from = from;
        this.to = to;
        this.heur = heur;
        this.work = work;
    }
    
    public void setViewer(MapViewer viewer) {
        this.viewer = viewer;
    }
    
    public void setWatcher(DistanceWatcher watcher) {
        this.watcher = watcher;
    }
    
    /** Carry out the search 
     * 
     * This method can run in a separate thread, so that the search
     * can be animated by a MapViewer.
     * */
    public synchronized float search() {
        for (Enumeration e = map.getTowns(); e.hasMoreElements(); ) {
            Town t = (Town) e.nextElement();
            t.setData(null);
        }           
        
        ActiveTown source = getActive(from);
        source.dist = 0.0f;
        source.known = true;
        source.prev = source;
        source.backlink = null;
        visitNeighbours(source);
        ActiveTown goal = getActive(to);
        
        // Loop until the shortest path to the finishing town is known
        while (!goal.known) {
            // Find the next town to visit
            ActiveTown t = findMin();
            if (t == null) {
                // Oops! No towns left to visit: that means there's no
                // feasible path
                return INF; 
            }
            
            // Mark the town as known
            t.known = true;
            if (work) {
                paint(t.backlink, Color.blue);
                if (t.town != to) paint(t, Color.white);
                if (watcher != null) watcher.update(t.dist);
            }
            
            // Update links for its neighbours
            visitNeighbours(t);
            
            try { 
                // Add 1 nanosecond to make sure timeout is nonzero,
                // since zero means infinity in Java-land!
                wait(delay, 1); 
            }
            catch (InterruptedException e) {};
        }

        // Paint the shortest path
        for (ActiveTown t = goal; t != source; t = t.prev)
            paint(t.backlink, Color.yellow);
        
        // Return the distance
        if (heur)
            return goal.dist + from.distance(to);
        else
            return goal.dist;
    }
    
    /** Update links for the neighbours of a town */
    private void visitNeighbours(ActiveTown t) {
        // Loop over all neighbours
        for (Enumeration z = t.town.getArcs(); z.hasMoreElements(); ) {
            Arc a = (Arc) z.nextElement();
            ActiveTown u = getActive(a.getDest());
            
            // Compute distance of the path that reaches u via t
            float d = t.dist + a.getRoad().length;
            
            if (heur) {
                // Heuristic: reduce d by the distance made good
                // towards the destination
                d -= t.town.distance(to) - u.town.distance(to);
            }
            
            if (u.prev == null) {
                // First time we have seen u
                u.dist = d;
                u.backlink = a.getRoad();
                u.prev = t;
                if (work) {
                    if (u.town != to) paint(u, Color.gray);
                    paint(a.getRoad(), Color.green);
                }
            }
            else if (d < u.dist) {
                // The route via t improves on a previous path
                if (work) {
                    paint(u.backlink, Color.magenta);
                    paint(a.getRoad(), Color.green);
                }
                u.dist = d;
                u.backlink = a.getRoad();
                u.prev = t;
            }
        }
    }
    
    /** Change the colour of a town and notify the viewer */
    private void paint(ActiveTown t, Color color) {
        t.town.setColor(color);
        if (viewer != null) 
            viewer.updateTown(t.town);
    }
    
    /** Change the colour of a road and notify the viewer */
    private void paint(Road r, Color color) {
        r.setColor(color);
        if (viewer != null)
            viewer.updateRoad(r);
    }
    
    /** Set the delay between steps.
     * 
     * This method is marked as synchronized to prevent interference
     * with the search method -- although in practice the interference
     * would be harmless anyway
     */
    public synchronized void setDelay(long delay) {
        this.delay = delay;
    }

    /** Find the next town to visit */
    private ActiveTown findMin() {
        // The desired town is the nearest among 
        // those not yet marked as known
        
        /* It would be good to use a priority queue here,
         * but for small maps, linear search is enough */
        ActiveTown min = null;
        for (Enumeration e = map.getTowns(); e.hasMoreElements();) {
            Town t = (Town) e.nextElement();
            ActiveTown a = (ActiveTown) t.getData();
            if (a != null && !a.known && a.backlink != null
                    && (min == null || a.dist < min.dist))
                min = a;
        }
        return min;
    }
    
    private ActiveTown getActive(Town t) {
        ActiveTown a = (ActiveTown) t.getData();
        if (a == null) {
            a = new ActiveTown(t);
            t.setData(a);
        }
        return a;
    }
}

/** Data about a town during a search. */
class ActiveTown {
    // These data members are public, because ActiveTowns are used only
    // in PathFinder, and accessor methods would be too clumsy.
    
    public Town town;
        // The town this data is about
    public float dist = PathFinder.INF;
        // Best known distance to this town
    public boolean known = false;
        // Whether shortest path to this town is known
    public Road backlink = null;
        // Final road on current best path
    public ActiveTown prev = null;
        // Previous town on current best path
    
    public ActiveTown(Town town) {
        this.town = town;
    }
}

interface DistanceWatcher {
    void update(float dist);
}