
package neosim.iface;

import java.util.Vector;
import java.util.Hashtable;

/** A neosim entity. All user entities implement this interface.
 * @author Fred Howell
 * @date Updated Nov 2000
 */
public interface Entity extends  java.lang.Comparable { //java.io.Serializable,

    /** User entities can extend this method 
     */
    public void init(Population p, int index);

    /** User entities MUST extend this method : deal with events in el up
     * to time t, and advance the entity's time to at least t and return the time 
     * actually advanced to (which may be later than t, as long as the state can be 
     * rewound back to t and no new events are generated after t). If the 
     * event type of incoming events is not recognised, the
     * method defaultEventHandler() should be called.
     * @param el The (possibly empty) event list in timestamp order
     * @param t The time to advance to
     * @return the time actually advanced to (must be >= t)
     * @see #defaultEventHandler
     */
    public Time handleEvents(EventList el, Time t);

    /** This method is called when simulation time
     * is rest to zero. User entities can extend this method to perform
     * a local reset.
     */
    public void reset();

    /** This method is called when the entity is deleted.
     * User entities can extend this to remove any locally
     * allocated storage.
     */
    public void destroy();
 

    /** This method sends an event from output port c.
     * @param c The output port to send the event to.
     * @param e The event to send. Its timestamp should
     * be the <em>sending time</em>
     */
    public void sendEvent( short c, Event e );//PortID c


    /** This method sends an event <b>to</b> input port p of entity deste.   
     * @param deste The destination entity
     * @param destp The destination input port
     * @param e The event to send. Its timestamp should
     * be the <em>sending time</em>
     */
    public void sendEvent( int deste, short destp, Event e );//EntityID deste, PortID destp//PortID destp


    /** This method deals with built in events, used for querying
     * and updating entity state.
     * @param e The event to handle
     */
    public void defaultEventHandler( Event e );

    /** Return the population this entity was created in.
     * @return the population this entity was created in.
     */
    public Population getPopulation() ;

    /** Return the index of this entity in its population
     * @return the index of this entity in its population
     */
    public int getIndex() ;

    /** Return the class of this entity.
     * @return the class of this entity.
     */
    public EntityClass getEntityClass();

    /** Return the ID of this entity.
     * @return the ID of this entity.
     */
    public int getID();//EntityID

    /** Return the name of this entity.
     * @return the name of this entity.
     */
    public String getName();

    /**  Return the top level population tree.
     * @param t the time of the request (needed as the population tree
     * changes with time)
     * @return the top level population tree
     *
     */
    public PopulationTree getPopulationTree(Time t);

    /** Return the set of user defined classes.
     * @return the set of user classes.
     *
     */
    public NeosimClasses  getNeosimClasses();
 
    /** Look up an EntityClass object given a name,
     * e.g. "CompartmentalEntity"
     * @param name The fully qualified name of the class
     * @return the entity class, if found, or null if not.
     */
    public EntityClass    lookupEntityClass( String name );

    /** Look up an Entity handle given a name,
     * e.g. "cortex.layer1.pyramidal1234"
     * @param t The time of the request
     * @param name The fully qualified name of the entity
     * @return the entity handle, if found, or null if not.
     */
    public int   lookupEntityID( Time t, String name );//EntityID return

    /** Look up a Population given a name,
     * e.g. "cortex.layer1"
     * @param t The time of the request
     * @param name The fully qualified name of the population
     * @return the population if found, or null if not.
     */
    public Population     lookupPopulation( Time t, String name );
 

    /** Build a population at a given time 
     * @param t the time to build the population (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to build
     * @param c the callback to call when the population has been built.
     */
    public  void   construct ( Time t, PopulationBuilder p, Callback c);

    /** Destroy a population at a given time. Calls every member's destroy() method.
     * @param t the time to build the population (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to destroy
     * @param c the callback to call when the population has been destroyed	
     */
    public void         destroy ( Time t, Population p, Callback c );

    /** Make connections at a given time
     * @param t the time to make the connections (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the projection to make
     * @param c the callback to call when the connections have been made
     */
    public void         connect ( Time t, Projection p, Callback c );

    /** Delete connections at a given time
     * @param t the time to delete the connections (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to destroy
     * @param c the callback to call when the population has been destroyed	
     */
    public void         disconnect ( Time t, Projection p, Callback c );

    /** Update another entity at a given time
     * @param t the time to update the entity (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param e the handle of the entity to update
     * @param eu the update to perform	
     * @param c the callback to call when the update has been done
     */
    public void         updateEntity( Time t, int e, EntityUpdate eu, Callback c );//EntityID e

    /** Query another entity at a given time
     * @param t the time to query the entity (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param e the handle of the entity to query
     * @param eq the query to perform	
     * @param c the callback to call with the results of the query
     */
    public void         queryEntity( Time t, int e, EntityQuery eq, Callback c );//EntityID e
    
    /** Update a population of entities at a given time
     * @param t the time to update the entity (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to update
     * @param eu the update to perform	
     * @param c the callback to call when the update has been done
     */
    public void         updatePopulation( Time t, Population p, EntityUpdate eu, Callback c );

    /** Query a population of entities at a given time
     * @param t the time to query the population(<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to query
     * @param eq the query to perform	
     * @param c the callback to call with the results of the query. The results will
     * be in the form of a vector of objects, indexed by entities' index in the population.
     */
    public void         queryPopulation( Time t, Population p, EntityQuery eq, Callback c );

    /** Query output connections of a port 
     * @param t the time to query the connections(<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the output port
     * @param cq the query to run on each output connection
     * @param c the callback to call with the results of the query. The results will
     * be in the form of a vector of objects, ordered by entity handle/input connection id.
     */
    public void queryOutputConnections( Time t, short p, ConnectionQuery cq, Callback cb );// PortID p


    /** Add an output port 
     * @param vec the vector of EventClass events the port generates
     * @return the id of the output port (from 1..#output ports).
     */
    short  addOutPort(Vector vec);//PortID

    /** Add an output port 
     * @param ec the EventClass the port generates
     * @return the id of the output port (from 1..#output ports).
     */
     short addOutPort(EventClass ec);//PortID

    /** Add an input port 
     * @param ec the EventClass the port accepts
     */
     short addInPort(EventClass ec);//PortID

    /** Add an input port 
     * @param vec the vector of EventClass the port accepts
     * @return the id of the input port (from 1..#input ports).
     */
    short   addInPort (Vector  vec);//PortID

    /** Return the class of events accepted by an input port
     * @param pid the input port id
     * @return the vector of EventClass 
     */
    Vector getClassesOfInPort(short  pid);//PortID

    /** Return the class of events produced by an output port
     * @param pid the output port id
     * @return the classes of events produced
     */
    Vector   getClassesOfOutPort(short  pid);//PortID

    /** Return the number of input connections on a port
     * @param inport the input port id
     * @return the number of connections
     */
    int  getNumInConnections(short  inport);//PortID

    /** Return the number of output connections from a port
     * @param outport the output port id
     * @return  the number of connections
     */
    int  getNumOutConnections(short  outport);//PortID

    /** Return the input connections to a port
     * @param outport the output port id
     * @return  the vector of connections
     */
    int[][] getInConnections(short  inport) ;//PortID
    
    /** Add an output connection from an output port to a given dest
     * @param outport the output port id
     * @param delay the delay on the connection
     * @param dest the destination entity
     * @param destinport the destination port id
     */
    void  addOutConnection(short outport, long delay, // PortID outport
			   int dest, short destinport );//EntityID dest//PortID destinport
     /** Remove an output connection from an output port to a given dest
     * @param outport the output port id
     * @param delay the delay on the connection
     * @param dest the destination entity
     * @param destinport the destination port id
     */
    void  removeOutConnection(short outport, long delay, //PortID outport
			   int dest, short destinport );//EntityID dest, PortID destinport

    /** Delete an output connection from an output port to a given dest
     * @param outport the output port id
     * @param dest the destination entity
     * @param destinport the destination port id
     */
    void  deleteOutConnection(short outport, int dest, //PortID outport, EntityID dest,//PortID outport
			      short destinport);//PortID destinport

    /** Add an input connection
     * @param c the input connection to add
     */
     int addInConnection(Connection c);//ConnectionID return


    /** Set additional connection parameters at src end 
     * Called by ParamConnectionSpec.setSrcParams
     */
    void setSrcConnectionParams( Connection c,  short srcpid, Hashtable params);//PortID srcpid

    /** Set additional connection parameters at dst end 
     * Called by ParamConnectionSpec.setDstParams
     */
    void setDstConnectionParams( Connection c, int destcid, Hashtable params );//ConnectionID destcid

    /** Delete an input connection
     * @param cid the input connection to delete
     */
    void  deleteInConnection(Connection c) ;//ConnectionID cid

    /** Delete an input port   
     * @param inport the input port id
     */
    void  deleteInPort(short inport) ;//PortID

    /** Delete an output port   
     * @param inport the input port id
     */
    void deleteOutPort(short outport);//PortID

    /** Return the list of incoming events */
    public EventList getEventList();
  
    /** Zero delay on system output port */
    void zeroOutputDelay();

    /** Set output delay to its previous value */
    void         resetOutputDelay();

    /** Get the minimum output delay */
    Time         getMinOutputDelay() ;

    /** Set the minimum output delay */
    void         setMinOutputDelay(Time t);

    /** Get the right window time ( local time + o/p delay ) */
    Time         getRightWindowTime();

    /** Get the current local time */
    Time         getLocalTime();

    /** Write a trace message */
    void trace(String msg);
}
