
package neosim.iface;

import java.util.Vector;
import neosim.kernel.SimKernel;

/** A generic script reader.   This has a bootstrap function
 * which takes over the System.main() thread and calls
 * methods to create new entities. 
 * @author Fred Howell
 * @version 1.0
 * @date Updated Oct 2000 to include initKernel method.
 */
public interface ScriptReader {

    /** Derived versions should override this method. 
     */
    public void bootstrap(String args[]);

    /** Initialise kernel - 
     * defaults (seq/thr/rmi) provided by ScriptReaderImpl
     */
    public SimKernel initKernel(String args[]);

    /** Initialise the simulation - called automatically
     */
    public void initSim();

    /** Run the simulation until the given time
     * @param t the time to run until.
     */
    public void runSim(Time t);
    
    /** Reset the simulation to time 0
     */
    public void reset();

    /** Terminate the kernel.
     */
    public void terminate();

    /** Construct a population of entities
     * @param p the population to build
     */
    public void construct ( PopulationBuilder p);

    /** Destroy a population of entities
     * @param p the population to destroy
     */
    public void destroy ( Population p );

    /** Connect up a projection
     * @param p the projection to make
     */
    public void connect ( Projection p );

    /** Disconnect a projection
     * @param p the projection to break
     */
    public void disconnect ( Projection p );


    /** Perform a generic update on an entity
     * @param e the id of the entity to update
     * @param eu the update to perform
     */   
    public void   updateEntity( int e, EntityUpdate eu );//EntityID


    /** Perform an query on an entity
     * @param e the id of the entity to update
     * @param eq the query to perform
     * @return the result of the query (as an Object)
     */       
    public Object queryEntity(  int e, EntityQuery eq );//EntityID

    /** Perform an update on a population
     * @param p the population to query
     * @param eu the update to perform
     */       
    public void   updatePopulation( Population p, EntityUpdate eu );

    /** Perform an query on a population
     * @param p the population to query
     * @param eq the query to perform
     * @return the result of the query (as an Object)
     */       
    public Vector queryPopulation(  Population p, EntityQuery eq );

    /** Perform an query on the output connections of an entity
     * @param e the source entity
     * @param p the source output port ID
     * @param cq the query to perform
     * @return the result of the query (as a Vector of Objects)
     */       
    public Vector queryOutputConnections( int e, short p, ConnectionQuery cq );//EntityID e//PortID p

    /** Return the root of the populations
     * @return the root of the populations
     */
    public PopulationTree getPopulationTree();

    /** Return the registry of classes
     * @return the registry of classes
     */
    public NeosimClasses  getNeosimClasses(); 

    /** Lookup an entity class given its name
     * @param name the name of the class
     * @return the entity class
     */
    public EntityClass    lookupEntityClass( String name );

    /** Lookup an entity ID given its name
     * @param name the name of the entity
     * @return the entity ID
     */
    public int       lookupEntityID( String name );//EntityID

     /** Lookup a Population given its name
     * @param name the name of the population
     * @return the population
     */   
    public Population     lookupPopulation( String name );
}
