package neosim.implementation;

import neosim.kernel.*;
import neosim.iface.*;
import java.util.Vector;

import neosim.seqkernel.*;
import neosim.thrkernel.*;
import neosim.rmikernel.*;

/** A generic script reader.   This has a bootstrap function
 * which takes over the System.main() thread and calls
 * methods to create new entities. 
 */

public class ScriptReaderImpl implements ScriptReader {

    Kernel sk;
    Time t = new Time(0);
    Callback cb;
    Semaphore sem = new Semaphore(0);

    boolean debug = false;

    public ScriptReaderImpl() {
	if (debug)  {
	    System.out.println("ScriptReader constructor "+this);
	}
	cb = new ScriptCallback( );
	SimSystem.setScriptReader( this );      
    }

    /** This method initialises the kernel variable */
    public final void initKernel() {
      sk = SimSystem.getKernel();
      if (debug)  {     
	  System.out.println("Initing scriptreaderimpl kernel" + sk);
      }
    }
	
    /** Derived versions should override this method. 
     */
    public final void bootstrap(String args[]) {
	//	System.out.println("The generic script reader with args "+args[0]);
	sk = SimSystem.getKernel();
	if (sk==null) { 
	    System.out.println("Error - kernel null in script reader bootstrap");
	}
    }

    protected boolean cons=true;

    /** Returns whether this processor needs to build the
     * simulation - usually only one has the script reader
     * bootstrap() method called.
     */
    public final boolean needToConstruct() { return cons; }
    protected final static int atoi(String s) { return (Integer.valueOf(s)).intValue(); }


    /** Initialise kernel - to one of seq/thr/rmi.
     * Each processor in a multiprocessor machine calls
     * this to initialise the local kernel.
     *
     * @param args the command line arguments
     */
    public final SimKernel initKernel(String args[]) {	
 	/* Initialise either seq/thr/rmi kernel */
	int argc = 0;
        cons = false; 
//	SimKernel sk;
        if (args.length < 1) {
            System.out.println("usage: java Basic {seq | thr <nthreads> | rmi [-c] <server> <user> <nprocs> <nthreads>}");
	    return null;
        } else {
            //if(sk!=null)
            //    sk.reset();//JAG
	    if (args[argc].equals("seq")) {
		sk = new SeqKernel();
		cons = true;
		argc++; 		
	    } else {
		if (args[argc].equals("thr")) {
		    argc++;
		    sk = new ThrKernel(atoi(args[argc]));
		    cons = true;
		    argc++; 			    
		} else {
		    if (args[argc].equals("rmi")) {
			argc++;
			if (args[argc].equals("-c")) { cons = true; argc++; }
			sk = new RmiKernel(args[argc],args[argc+1],atoi(args[argc+2]),atoi(args[argc+3]));
			argc+=4;
		    } else {
			System.out.println("usage: java Basic {seq | thr <nthreads> | rmi [-c] <server> <user> <nprocs> <nthreads>}");
			return null;
		    }
		}
            }
	}
	sk.init();
	this.sk = sk; // added fwh oct 2001
	return (SimKernel)sk;
    }

    public final void destroyKernel() {
	if (sk!=null) {
	    sk.destroy(); // This issues a command everywhere to destroy
	} else {
	    System.out.println("ScriptReaderImpl:destroy Kernel - error kernel null");
	}
    }

    public final void initSim() {
        if (debug) System.out.println("ScriptReaderImpl:initSim");
        sk = SimSystem.getKernel();
	if (sk!=null) {
	  sk.initSim();
	} else System.out.println("ScriptReaderImpl:initSim error - kernel null");
        if (debug) System.out.println("ScriptReaderImpl:initSim done");
    }

    public final void runSim(Time rt) {
	//	System.out.println("Script Reader: run sim "+rt);
	sk.runSim(rt);
	t.set(rt);
    }

    public final void reset() {
	sk.reset();
	t.set(0);
    }
    
    private boolean terminated = false;
    public final void terminate() {
	if (!terminated) {
	    sk.terminate();
	    terminated = true;
	} else 
	    System.out.println("ScriptReaderImpl Error - terminate() called twice");
    }

    public final void construct ( PopulationBuilder p) {
	//	System.out.println("SRImpl construct");
	sk.construct(EntityID.SYSTEM_ENTITY, t,p,cb);      
	//	System.out.println("SRImpl construct called, pausing");	
	pauseme();
	//	System.out.println("SRImpl construct called, releasing pause");	
    }

    public final void destroy ( Population p ) {
	sk.destroy(EntityID.SYSTEM_ENTITY,t,p,cb);
	pauseme();
    }

    public final void connect ( Projection p ) {
	sk.connect(EntityID.SYSTEM_ENTITY,t,p,cb);
	pauseme();
    }

    public final void disconnect ( Projection p ) {
	sk.disconnect(EntityID.SYSTEM_ENTITY,t,p, cb);
	pauseme();
    }

    public final void  updateEntity( int e, EntityUpdate eu ) {//EntityID e
	sk.updateEntity(EntityID.SYSTEM_ENTITY,t, e, eu, cb );
	pauseme();
    }

    public final Object queryEntity(  int e, EntityQuery eq ) {//EntityID e
	sk.queryEntity(EntityID.SYSTEM_ENTITY,t, e, eq, cb );	 
	pauseme();
	return answer;
    }

    public final void   updatePopulation( Population p, EntityUpdate eu ) {
	sk.updatePopulation (EntityID.SYSTEM_ENTITY,t, p, eu, cb ); 
	pauseme();
    }

    public final Vector queryPopulation(  Population p, EntityQuery eq ) {
	sk.queryPopulation (EntityID.SYSTEM_ENTITY,t, p, eq, cb ); 
	pauseme();
	return (Vector)answer;
    }

    public final Vector queryOutputConnections( int e, short p, ConnectionQuery cq ) {//EntityID e// PortID p
	sk.queryOutputConnections (EntityID.SYSTEM_ENTITY,t, e, p, cq, cb ); 
	pauseme();
	return (Vector)answer;
    }


    public final PopulationTree getPopulationTree() {	
	if (sk==null) { System.out.println("SK is NULL!"); }
	return sk.getPopulationTree(t);
    }

    public final NeosimClasses  getNeosimClasses() {
	return sk.getNeosimClasses();
    }

    public final EntityClass    lookupEntityClass( String name ) {
	return sk.lookupEntityClass( name );
    }

    public final int   lookupEntityID( String name ) {//EntityID
	return sk.lookupEntityID( t, name );
    }

    public final Population     lookupPopulation( String name ) {
	return sk.lookupPopulation( t, name );	
    }


    /** Test implementation of send SysEvent accessible from script
     * intended for adding simulation modules
     */
    public final void sendSysEvent( SysEntEvent e ) {
	//	System.out.println("sendSysEvent starting");
	e.setCallback( cb );
	//	System.out.println("sendSysEvent about to send "+sk);
	sk.sendEvent( EntityID.SYSTEM_ENTITY, PortID.SYSTEM_PORT, e );

	//	System.out.println("sendSysEvent about to pause");
	pauseme();
    }

    Object answer;

    /** Wait for callback */
    public final Object pauseme() {
	// sk.processRequests()

	//	System.out.println("Script reader pausing pending callback...");

	sem.semWait(this);

	//	System.out.println("Script reader got callback...");
	return answer;
    }

    /** Called on callback */
    public final void restart(Object o) {
	answer = o; 
	//	System.out.println("Script reader signalling callback...");
	sem.semSignal();
    }
}
