
package neosim.kernel;
import neosim.iface.*;
import neosim.implementation.*;
import java.util.Vector;

/** The neosim kernel interface. Just visible internally to implementations of Entity.
 */
public interface Kernel {
    public void runSim(Time t);
    public void initSim();
    public void init();
    public void destroy();
    public void terminate();
    public void reset();
 
    /** This method sends an event on output port c.
     * @param handle The originating entity.
     * @param c The output port to send the event from.
     * @param e The event to send. Its timestamp should
     * be the <em>sending time</em>
     */
    public void sendEvent( int handle, short p, Event e );//EntityID handle//PortID p

    /** This method sends an event to an input port of the dest entity.
     * @param handle The originating entity.
     * @param srcp The output port to send the event from.
     * @param deste The destination entity
     * @param destp The destination input port 
     * @param e The event to send. Its timestamp should
     * be the <em>sending time</em>
     */
    public void sendEvent( int handle, short srcp, int deste, short destp, Event e );//EntityID handle,//EntityID deste// PortID srcp//PortID destp

    /** This method deals with built in events, used for querying
     * and updating entity state.
     * @param ent The entity the event arrived at
     * @param e The event to handle
     */
    public void defaultEventHandler( Entity ent, Event e );

 
    /**  Return the top level population tree.
     * @param t the time of the request (needed as the population tree
     * changes with time)
     * @return the top level population tree
     *
     */
    public PopulationTree getPopulationTree(Time t);

    /** Return the set of user defined classes.
     * @return the set of user classes.
     *
     */
    public NeosimClasses  getNeosimClasses();
 
    /** Look up an EntityClass object given a name,
     * e.g. "CompartmentalEntity"
     * @param name The fully qualified name of the class
     * @return the entity class, if found, or null if not.
     */
    public EntityClass    lookupEntityClass( String name );

    /** Look up an Entity handle given a name,
     * e.g. "cortex.layer1.pyramidal1234"
     * @param t The time of the request
     * @param name The fully qualified name of the entity
     * @return the entity handle, if found, or null if not.
     */
    public int   lookupEntityID( Time t, String name );//EntityID

    /** Look up a Population given a name,
     * e.g. "cortex.layer1"
     * @param t The time of the request
     * @param name The fully qualified name of the population
     * @return the population if found, or null if not.
     */
    public Population     lookupPopulation( Time t, String name );
 

    /** Build a population at a given time 
     * @param e the originating entity
     * @param t the time to build the population (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to build
     * @param c the callback to call when the population has been built.
     */
    public  void   construct ( int e, Time t, PopulationBuilder p, Callback c);//EntityID e

    /** Destroy a population at a given time. Calls every member's destroy() method.
     * @param e the originating entity
     * @param t the time to build the population (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to destroy
     * @param c the callback to call when the population has been destroyed	
     */
    public void         destroy ( int e, Time t, Population p, Callback c );//EntityID

    /** Make connections at a given time
     * @param e the originating entity
     * @param t the time to make the connections (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the projection to make
     * @param c the callback to call when the connections have been made
     */
    public void         connect ( int e, Time t, Projection p, Callback c );//EntityID

    /** Delete connections at a given time
     * @param e the originating entity
     * @param t the time to delete the connections (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to destroy
     * @param c the callback to call when the population has been destroyed	
     */
    public void         disconnect ( int e, Time t, Projection p, Callback c );//EntityID e

    /** Update another entity at a given time
     * @param e the originating entity
     * @param t the time to update the entity (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param d the handle of the entity to update
     * @param eu the update to perform	
     * @param c the callback to call when the update has been done
     */
    public void         updateEntity( int e,Time t, int d, EntityUpdate eu, Callback c );//EntityID e, //EntityID d

    /** Query another entity at a given time
     * @param e the originating entity
     * @param t the time to query the entity (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param d the handle of the entity to query
     * @param eq the query to perform	
     * @param c the callback to call with the results of the query
     */
    public void         queryEntity( int e,Time t, int d, EntityQuery eq, Callback c );//EntityID e, //EntityID d
    
    /** Update a population of entities at a given time
     * @param e the originating entity
     * @param t the time to update the entity (<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to update
     * @param eu the update to perform	
     * @param c the callback to call when the update has been done
     */
    public void         updatePopulation( int e, Time t, Population p, EntityUpdate eu, Callback c );//EntityID e

    /** Query a population of entities at a given time
     * @param e the originating entity
     * @param t the time to query the population(<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param p the population to query
     * @param eq the query to perform	
     * @param c the callback to call with the results of the query. The results will
     * be in the form of a vector of objects, indexed by entties' index in the population.
     */
    public void         queryPopulation( int e, Time t, Population p, EntityQuery eq, Callback c );//EntityID e


     /** Query output connections of a port 
     * @param oe the originating entity
     * @param t the time to query the connections(<b> must be >= localTime() + getMinOutputDelay()</b>)
     * @param srce the entity handle of the src entity to query
     * @param srcp the src output port
     * @param cq the query to run on each output connection
     * @param c the callback to call with the results of the query. The results will
     * be in the form of a vector of objects, ordered by entity handle/input connection id.
     */
    public void queryOutputConnections( int oe, Time t, int srce, short srcp, ConnectionQuery cq, Callback cb );//EntityID oe //EntityID srce//PortID srcp



    /** Return the number of input connections on a port
      * @param e the originating entity
      * @param inport the input port id
      * @return the number of connections
      */
    public int  getNumInConnections(int e, short inport);////EntityID e//PortID inport

    /** Return the input connections to a port
     * @param e the originating entity
     * @param outport the output port id
     * @return  the number of connections
     */
    Vector getInConnections(int e, short inport);//EntityID e//PortID inport

    /** Return processor where an entity is stored
     * @param e the entity handle
     * @return  the processor id
     */
    int getEntityProcessor(int e);//EntityID e

    /** Add an entity
     * @param e the entity to add
     */
    void addEntity(EntityImpl e);

    /** Write a trace message to the tracefile.
     * @param msg the message to write
     */
    void trace(String msg);
    /** Add a spike to the spike list
     */
    void addSpike(int handle, short p, long t);
    
     /** Exchange the
     */
    void exchangeSpikes();
    
   
}
