/**
 * The neosim PDES kernel.
 * 
 * @version     
 *      0.1 1st July 1998
 *      0.11 13th August 1998 - removed dead simjava stuff
 *      0.2 2nd Nov 1998 - RMI version
 *      0.3 Oct 2000 = Threads version
 *
 * @author      Fred Howell
 */

package neosim.kernel;
 
import neosim.iface.*;
import neosim.implementation.*;

import neosim.perf.*;

import java.text.NumberFormat;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * This is the system class which manages the simulation. All of
 * the members of this class are static, so there is no need to
 * create an instance of this class.
 * @version     1.0, 4 September 1996
 * @version     1.2, 14 July 1997
 * @author      Ross McNab, Fred Howell
 */

public final class SimSystem {
    //static neosim.kernel.DeadlockTracker tracker = new neosim.kernel.DeadlockTracker();
    static private boolean debug = false;
//JAG DANGEROUS OPTIMIZATION
    static private long TickTime = 0;
    static private boolean calculated = false;
    
    
    static private Vector entities;	// The current entity list
    static private Hashtable entityIdx; // Entity by handle
    static private ProcManager processors;// The current processor list
    static private ThreadManager threads;// The current thread list
 
    static private Time clock = new Time(0);   // Holds the current global sim time
    static private boolean running;	// Has the run() member been called yet
    static private Semaphore onestopped;
    static private SimOutput trcout;	// The output object for trace messages
    static private int trc_level;
    static private boolean auto_trace;	// Should we print auto trace messages?
    static private NumberFormat nf;
    
    
    // --- Extra stuff for parallel simulation ---
    static private SimKernel sk;
    static public  SimKernel getSk()  { return sk; }
    static public  SimKernel getKernel()  { return sk; }
    static public  void setKernel(SimKernel ski)  { sk = ski; }

    // --- Extra user info, used for callbacks.
    static ScriptReaderImpl sr;
    static public void setScriptReader(ScriptReaderImpl sri) { sr = sri; }
    static public ScriptReaderImpl getScriptReader() { return sr; }
   
    // --- Extra stuff for time windows. ---
    static private Entqueue runnable_entity_heap;
    static private Entqueue blocked_entity_heap;
    static private Time left_window = new Time(0), right_window=new Time(0), min_delay= new Time(0);
    static private Time endTime = new Time(0);
    
    // --- Extra system entity ---
    static private SystemEntity sysEnt;
  
    // --- Performance monitoring
    static private PerfMonitor procMon=null;
    static private boolean useRealTime=true;
    static private PerfMonitor entMon=null;
    static private double getRealTime() {
	return System.currentTimeMillis();
    }
    static public Time getSimTime() {
	return clock;
    }
    static double t0;
    // Get initial time.
    static private void initMonTime() {
	if (useRealTime) t0= System.currentTimeMillis(); else t0=0;
    }
    static public double getMonTime() {
	if (useRealTime) return System.currentTimeMillis()-t0; else return clock.s();
    }
    /* Performance monitoring commands */
    static public void monitorProcessors(PerfMonitor pmon, boolean realTime) {
	procMon = pmon; useRealTime = realTime;
	initMonTime();
    }
    /* Performance monitoring commands */
    static public void monitorEntities(PerfMonitor pmon, boolean realTime) {
 	entMon = pmon;  useRealTime = realTime;
	initMonTime();
    } 


    /** Initialise the system, this function does the job of a
     * constructor, and should be called at the start of any simulation
     * program. It comes in several flavours depending on what context
     * the simulation is running.<P>
     * This is the simplest, and should be used for standalone simulations
     * It sets up trace output to a file in the current directory called
     * `tracefile'
     */
    static public void initialise() {
	initialise(new SimOutfile("tracefile"), null);
    }
    
    /** Initialise the system, with a parallel kernel
     */
    static public void initialise(SimKernel skin, int pid) {
	if (debug) {	    
	    System.out.println("SimSystem initialising parallel processor "+pid);
	}
	sk = skin;
	initialise(new SimOutfile("tracefile"+pid), null);
    }
    
    /** This version of initialise() is for experienced users who
     * want to use the trace output to draw a graph or an animation
     * as part of the application.
     * @param out The object to be used for trace output
     * @param sim The thread the simulation is running under, (set to
     *            <tt>null</tt> if not applicable
     */
    static public void initialise(SimOutput out, Thread sim) {
	System.out.println("\n======= Neosim 1.2, Apr 2001 =========");
	entities   = new Vector();
	entityIdx  = new Hashtable();
	trcout     = out;
	trc_level  = 0xff;
	auto_trace = false;
	
	// Set the default number format
	nf = NumberFormat.getInstance();
	nf.setMaximumFractionDigits(4);
	nf.setMinimumFractionDigits(2);
	//	System.out.println("\n\n======= about to reset vars =========");
	
	resetVars();

	//	System.out.println("\n\n======= about to make system entity =========");

	// Create SysEntity, used for commands during simulation run.
	sysEnt = new SystemEntity(getSk());
	//	System.out.println("\n\n======= done initialise =========");
    }
    
    /** Resets variables for next run
     */
    static public void resetVars() {
	processors = new ProcManager(); 
	clock      = new Time(0.0);
	running    = false;
	onestopped = new Semaphore(0);
	trcout.initialise();

	// CTW stuff
	blocked_entity_heap  = new Entqueue();
	runnable_entity_heap = new Entqueue();
	left_window          = new Time(0.0);
	//	min_delay            = new Time(5.0); // getMinDelay()
	min_delay            = calcMinOutputDelay(); // **** NOT PARALLEL ***
	right_window         = new Time(); 
	right_window.t = left_window.t + min_delay.t;	

	// PerfMon stuff
	if (procMon!=null) { procMon.resetEvents(); initMonTime(); }
	if (entMon!=null)  { entMon.resetEvents();  initMonTime(); }
        
        //JAG
        if(threads!=null)
            threads.killThreads();
        
        //JAG DANGEROUS OPTIMIZATION
         calculated=false;       
        
    }

    /** Returns the ProcManager
     */
    static public ProcManager getProcManager() { return processors; }

    /** Returns the System Entity
     */
    static public SystemEntity getSysEntity() { return sysEnt; }
    
    /** Returns the number format used for generating
     * times in trace lines
     */
    static public NumberFormat getNumberFormat() { return nf; }
    
    /** Get the current simulation time.
     * @return The simulation time
     */
    static public Time clock() { return clock; }
    
    /** Get the current number of entities in the simulation
     * @return A count of entities
     */
    static public int getNumEntities() { return entities.size(); }
    
    /** Get the current trace level (initially <tt>0xff</tt>), which
     * controls trace output.
     * @return The trace level flags
     */
    static public int getTrcLevel() { return trc_level; }
    
    /** Find an entity by its id number.
     * @param id The entity's local index
     * @return A reference to the entity, or null if it could not be found
     */
    static public Entity getEntity(int i) {
	return (Entity)entities.elementAt(i);
    }

    /** Find an entity by its handle.
     * @param h The entity's unique handle
     * @return A reference to the entity, or null if it could not be found
     */
    static public Entity getEntity(Integer h) {//EntityID h
	return (Entity)entityIdx.get( h );
    }

    /** Find an entity by its name.
     * @param name The entity's name
     * @return A reference to the entity, or null if it could not be found
     */
    static public Entity getEntity(String name) {
	Enumeration e;
	Entity  ent, found = null;
	
	for (e = entities.elements(); e.hasMoreElements();) {
	    ent = (Entity)e.nextElement();
	    if(name.compareTo(ent.getName()) == 0)
		found = ent;
	}
	if(found == null)
	    System.out.println("SimSystem: could not find entity "+name);
	return found;
    }
 
    /** Calculate minimum output delay for all entities
     * *** Note - this isn't sufficient for the parallel version ***
     * @return minimum output delay, or -1.0 if none.
     */
    static public Time calcMinOutputDelay() {
	//	System.out.println("calc min o/p delay");
        //new Exception().printStackTrace();
	Time min = new Time(); min.t=-1;
	Entity ent;
	int i;
	for (Enumeration e = entities.elements(); e.hasMoreElements();) {
	    ent = (Entity)e.nextElement();
	    Time ed = ent.getMinOutputDelay();
	    //	    System.out.println("Entity "+ent.getName()+"  min delay "+ed.s());
	    if ((min.t==-1) || ((ed.t!=-1) && ed.t < min.t)){
		min = ed;
	    }
	}
	return min;
    }

    /** Set delay for all entities to d
     */
    static void setMinOutputDelay(Time d) {
	for (Enumeration e = entities.elements(); e.hasMoreElements();) {
	    Entity ent = (Entity)e.nextElement();
	    ent.setMinOutputDelay(d);
	}
    }

    /** Find out an entities unique id number from its name.
     * @param name The entity's name
     * @return The entity's unique id number, or -1 if it could not be found
     */
    static public int getEntityId(String name) {
	return entities.indexOf(getEntity(name));
    }

    /** Set the trace level flags which control trace output.
     * @param level The new flags
     */
    static public void setTrcLevel(int level) { trc_level = level; }

    /** Switch the auto trace messages on and off.
     * @param on If <tt>true</tt> then the messages are switched on
     */
    static public void setAutoTrace(boolean on) { auto_trace = on; }

    /** Add a new entity to the simulation.
     * This is now done automatically in the Entity constructor,
     * so there is no need to call this explicitly.
     * @param e A reference to the new entity
     */
    static public void add(Entity e) {
	if (running()) {
	    /* Post an event to make this entity */
	    // *** Disabled for now ***
	} else {	    
	    entities.addElement(e);
	    entityIdx.put( new Integer(e.getID()), e );
	    //	    System.out.println("Adding entity "+e.getID().id);
	}
    }
 
    /** Start the simulation running. This should be called after
     * all the entities have been setup and added, and their ports linked
     * Phase 1.
     */
    static public void runStart(Time startT) {
	Enumeration e;
	Entity  ent;
	int i;
	
	running = true;
	clock = startT;
    left_window = startT;
	
	//	System.out.println("SimSystem: Starting entities");
	for (e = entities.elements(); e.hasMoreElements();) {
	    ent = (Entity)e.nextElement();
            //neosim.kernel.Deadlock_Message ecp=new neosim.kernel.Deadlock_Message();
            //neosim.kernel.DeadlockTracker.LockDeadlockTracker(e,ecp);             
	    runnable_entity_heap.add(ent);
            //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp);
	}
    }
    
  
    /** Run one tick of the simulation - parallel version.
     * Called from SimProcessor objects.
     * Grab one entity, run it as far as it can go.
     * Simple version - one SimProcessor per node.
     * But multiple Threads.
     * Algorithm:
     *   get min rw time, globally, exchg all events.
     *   advance safe time to this.
     *   select one or more entities to advance to this safe time, until rw is advancable, or
     *      all local entities done.
     */
    static public boolean runTickPar(int pid, SimProcessor p) {
	Entity ent;

	if (debug) {	System.out.println("runTickPar"); }

        //JAG DANGEROUS OPTIMIZATION
	//procState(pid, "SYNC");
	
       if(!calculated){
	// Get min rw across all local ents  
            //neosim.kernel.Deadlock_Message ecp=new neosim.kernel.Deadlock_Message();
            //neosim.kernel.DeadlockTracker.LockDeadlockTracker(p,ecp);              
            Time minrw = runnable_entity_heap.getMinrw(); 
            //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp);
            if (debug) {
                //neosim.kernel.Deadlock_Message ecp2=new neosim.kernel.Deadlock_Message();
                //neosim.kernel.DeadlockTracker.LockDeadlockTracker(p,ecp2);   
                System.out.println("Clock "+clock.s()+
			       "  Entities:"+runnable_entity_heap.size()+
			       " Local min rw = "+minrw.s());
                //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp2);   
            }                
            Time globMin = minrw;
            Vector alltimes;
            try {
                alltimes = sk.allgather( minrw );	    
                for (int i=0; i<alltimes.size(); i++) {
                    if (globMin.t > ((Time)(alltimes.elementAt(i))).t) {
                        globMin = (Time)(alltimes.elementAt(i));
                    }
                }	
            } catch (Exception e ) { System.out.println(e); }



            if (debug) {	System.out.println("Clock "+clock.s()+"  Global min rw = "+globMin.s()); }
            calculated=true;
            TickTime=globMin.t;
       }
	 //Get min rw across all processors
	//	Time safe_time = sk.getMinTime();
 	Time safe_time = new Time(0);
	
        //JAG DANGEROUS OPTIMIZATION
        //safe_time.t = globMin.t;  
        safe_time.t=clock.t+TickTime;
        
        //System.out.println("TICK till "+ safe_time.t+" Clock="+clock.t+" TickTime="+TickTime);
        
	/** Only run til 'endTime' */
	if (safe_time.t > endTime.t) { 
	    safe_time.t = endTime.t;
	}
	

	//	p.setRightWindowTime( safe_time );
        //neosim.kernel.Deadlock_Message ecp3=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp3);     
	advanceClock(safe_time);		// Advancing time.
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp3);
        
	boolean FlushMeasurementToo=(clock.t >= endTime.t); 
            
	// Signal this processor as running til safe time
	sk.setRunning(safe_time);
	if (debug) { System.out.println(pid+" ======================================== running until "+safe_time.s()); }
	sk.flush(FlushMeasurementToo);
//	sk.barrier();
	if (debug) {	System.out.println(pid+" -------------- All outstanding events will have arrived now"); }

        //JAG DANGEROUS OPTIMIZATION
	//procState(pid,"RUN");
	

	// Set threads going on the entity heap
	threads.runThreads( safe_time );

	if (false) {	
	    System.out.println("  SimSystem runTickPar: Setting stopped --- why?"); 
	}
	
        //JAG DANGEROUS OPTIMIZATION
	//procState(pid,"STOP");

	// ** Commented out sk.stopped! don't actually want to send a message to the effect! **
	//   sk.setStopped();
        //neosim.kernel.Deadlock_Message ecp5=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp5); 
	int sz = blocked_entity_heap.size();
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp5);
	if (debug) {	
	    System.out.println("  SimSystem runTickPar: moving "+sz+" ents to runnable heap"); 
	}
	
	for (int i=0; i<sz; i++) {
            //neosim.kernel.Deadlock_Message ecp2=new neosim.kernel.Deadlock_Message();
            //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp2);             
	    runnable_entity_heap.add(blocked_entity_heap.pop());
            //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp2);
	}
	if (debug) {	
	    System.out.println("  SimSystem runTickPar: done that"); 
	}
	
	if (clock.t >= endTime.t) {
	    running = false;
	    if (debug) {	    System.out.println("SimSystem: Stopping"); }
	}
	return running;
    }

    /** runEntities - pull entities off runnable heap as fast as
     * possible, and call runEntityTickPar.
     * Called from SimThread.run()  (started by ThreadManager.runThreads) 
     */
    static public void runEntities(SimProcessor sp, Time safe_time ) {
	boolean more = true;
	while (more) {
            //neosim.kernel.Deadlock_Message ecp2=new neosim.kernel.Deadlock_Message();
            //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp2); 
	    Entity ent = runnable_entity_heap.pop();
            //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp2);
	    if (ent == null) { more = false ; }
	    else {
		Time lw = ent.getLocalTime();
		Time rw = ent.getRightWindowTime();	                     
		runEntityTickPar2(sp,ent,lw,rw, safe_time);
                //neosim.kernel.Deadlock_Message ecp3=new neosim.kernel.Deadlock_Message();
                //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp3);                 
		blocked_entity_heap.add(ent);
                //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp3);
	    } 
	} 
    }


    /** Run an entity for a tick - cut down version
     */
    static public void runEntityTickPar2(SimProcessor p, Entity ent,
					Time lw, Time rw, Time safe_time) {

  	procState( p.getID(), "RUN" );
	if (debug) System.out.println(" ... advancing entity "+ent+" to "+safe_time);
        //neosim.kernel.Deadlock_Message ecp3=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp3);        
        EventList list=ent.getEventList();
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp3);       
	Time advt = ent.handleEvents( list, safe_time );
	if (debug) System.out.println(" ... done advancing entity "+ent+" to "+safe_time);
  	procState( p.getID(), "SYNC" );
 	EntityImpl nei = (EntityImpl)ent;
	nei.setLocalTime( safe_time );
	nei.setProvLocalTime( advt ); 
    }
    
    /** Advance clock
     */
    static public synchronized void advanceClock(Time t) {
	//	trcout.println("Advance clock "+clock.s()+" to "+t.s());
	if (clock.t < t.t) clock = t;
    }


    /** Run an entity for a tick *** NB Not Synchronized! ***
     */
    static public void runEntityTickPar(SimProcessor p, Entity ent,
					Time lw, Time rw, Time safe_time) {

	if (debug) {	System.out.println("Run Entity Tick Par abt to call handle events, p="+p); }
        //neosim.kernel.Deadlock_Message ecp3=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp3);   
	processors.setRunning(p, safe_time);
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp3);
        
	if (debug) {	System.out.println("Run Entity Tick Par abt to call handle events"); }
        //neosim.kernel.Deadlock_Message ecp6=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(ent,ecp6);    
        EventList el=ent.getEventList();
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp6);
        
	Time advt = ent.handleEvents(el,safe_time);
	if (debug) {	System.out.println("Run Entity Tick Par called handle events - advanced to "+advt.s()); }

	EntityImpl nei = (EntityImpl)ent;
	nei.setLocalTime( safe_time );
	nei.setProvLocalTime( advt );
        
        //neosim.kernel.Deadlock_Message ecp4=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp4);  
	processors.setStopped(p);
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp4);
    }
    
    /** Stop the simulation
     */
    static public void runStop() {
	Enumeration e;
	Entity  ent;
	for (e = entities.elements(); e.hasMoreElements();) {
	    ent = (Entity)e.nextElement();
	    //	    ent.poison();
	}
	//trcout.println("Exiting SimSystem.run()");
	// trcout.flush();
    }

    /** Processor changed state
     */
    static void procState(int pid, String s) {
	if (procMon != null) { procMon.addEvent(pid, getMonTime(), s);  }
    }

    /** Initialise performance monitoring
     */
    static void initPerfMon() {
	/* Performance monitoring */
	if (procMon!=null) {
	    procMon.registerType("ProcState", "RUN SYNC SLEEP STOP");
	    for (int p=0; p<processors.size(); p++) {
		procMon.registerEntityState("Proc"+p, "ProcState");
	    }	    
	    procMon.resetEvents();
	}
	if (entMon!=null) {
	    entMon.registerType("EntState", "SEND RECV PROCESS IDLE GROW CONNECT OUTPUT");
	    for (Enumeration e = entities.elements(); e.hasMoreElements();) {
		Entity ent = (Entity)e.nextElement();
		entMon.registerEntityState(ent.getName(), "EntState"); 
		//		ent.setPerfMonitor(entMon);
	    }
	    entMon.resetEvents();
	}
    }

    
    /** Start the simulation running. This should be called after
     * all the entities have been setup and added, and their ports linked.
     */
    static public void runPar(int nprocs, int localthreads, int pid, Time t) {
	int i;
	endTime.t = clock.t + t.t; // this is a run for! paul 
	int localprocs = 1;
	if (debug) { 	System.out.println("RunPar started "+endTime.s()+" threads:"+localthreads+" procs:"+nprocs); }


	//	System.out.println("run start");

	runStart(clock);  /** Start at the current clock value */

	if (debug) { 
	    trcout.println(" *** ["+pid+"] SimSystem: Entering parallel main loop from "+
			   clock.s()+" until time "+endTime.s());
	    trcout.flush();
	}
 


	Time mind = calcMinOutputDelay();
	if (mind.t==-1)  { mind = new Time(1000.0); }
	if (mind.t==0)   { mind = new Time(10.0); }
	if (debug) { trcout.println(" *** min delay is "+mind.s()); }
	setMinOutputDelay(mind);
	
	SimProcessor proc;
	processors = new ProcManager();

	/** Just one SimProcessor locally */
	proc = new SimProcessor(0);
	proc.start();
	processors.addElement(proc);

	/** Create a number of threads to run entities 
	 * Note - the sequence is:
	 * SimProcessor calls SimSystem.runTickPar
	 *  which restarts the SimThreads
	 *  which call SimSystem.runEntities in parallel
	 */
	threads = new ThreadManager( proc, localthreads );

	initPerfMon();

	if (debug) { System.out.println(pid+" *** SimSystem: Waiting for all processors to startup"); }
	sk.barrier(); 
	
	// Wait 'till the local SimProcessor is up and ready
	for(i=0; i<localprocs; i++) { onestopped.p(sk); }
	
	if (debug) { System.out.println( pid+ " *** SimSystem: About to start "+nprocs+" processors, "+localthreads+" threads going."); }
	
	// Start local SimProcessor
	SimProcessor p;
	for(i=0; i<localprocs; i++) {
	    p = (SimProcessor)processors.elementAt(i);
	    p.restart();
	}
	if (debug) { System.out.println( pid+" *** SimSystem: Running..."); }

	
	// Wait for all to finish
	for(i=0; i<localprocs; i++) { onestopped.p(sk); }

	if (debug) { System.out.println( pid+" *** SimSystem: all procs stopped..."); }

	// Kill local threads
	threads.killThreads();
	
	if (debug) { System.out.println( pid+" *** SimSystem: All threads stopped - waiting on barrier"); }
	sk.barrier();

	if (debug) { System.out.println( pid+" *** SimSystem: barrier done"); }
	
	runStop();
	if (debug) { System.out.println( pid+" *** SimSystem: runstop called "); }

    }
    
    
    static boolean running() { return running; }
    static public SimOutput getTrcout() { return trcout; }
    
    // Called by an entity just before it become non-RUNNABLE
    static void paused() { onestopped.v(); }
    
    /** Trace message */
    static synchronized void trace(int src, String msg) {
	trcout.println("u:"+((Entity)entities.elementAt(src)).getName()+
		       " at "+clock+": "+msg);
    }
    
     /** Raw trace message */
    static synchronized void rawtrace(String msg) {
	trcout.println(msg);
    }
   /** processor trace file */
    static synchronized public void ptrace(int proc, String msg) {
	trcout.println("u:Processor"+proc+" at "+clock.s()+": "+msg);
	trcout.flush();
    }
    
    /** generate VRML */
    static public void genVrml(String vfile) {
	SimOutfile so = new SimOutfile(vfile);
	so.initialise();
	// Generate headers ...
	// Generate instances
	for (int i=0; i<entities.size(); i++) {
	    //	    ((Entity)entities.elementAt(i)).genVrmlInst(so);
	}
	so.close();
    }    

    /** Reset a simulation. Sets all entities back to time 0.0
     * Only call when the simulation has stopped.
     */
    static public void reset() {
	for (int i=0; i<entities.size(); i++) {
	    ((Entity)entities.elementAt(i)).reset();
	}
	resetVars();
    }

    /** End a simulation. close the trace file
     */
    static public void terminate() {
	if (procMon!=null) { procMon.finished(); }
	if (entMon!=null) { entMon.finished(); }
	trcout.close();
 	for (int i=0; i<entities.size(); i++) {
	    //	    ((Entity)entities.elementAt(i)).terminate();
	}
   }

    /** Destroy a simulation. Removes all entities.
     */
    static public void destroy() {
	if (procMon!=null) { procMon.resetAll(); }
	if (entMon!=null)  { entMon.resetAll(); }
	for (int i=0; i<entities.size(); i++) {
	    ((Entity)entities.elementAt(i)).destroy();
	}
    }

    /** Start the simulation running [ SMP version ]. This should be called after
     * all the entities have been setup and added, and their ports linked
     */
    static public void runSmp(int nprocs) {
	int i;
	if (debug) {	System.out.println("SimSystem: Entering SMP main loop"); }
	runStart(clock);
	
	// Start N processors
	SimProcessor proc;
	for (int p=0; p<nprocs; p++) {
	    proc = new SimProcessor(p);
	    proc.start();
	    processors.addElement(proc);
	}

	if (debug) { System.out.println("SimSystem: Waiting for processors to startup"); }
	
	// Wait 'till they're all up and ready
	for(i=0; i<nprocs; i++) { onestopped.p(sk); }
	
	if (debug) { System.out.println("SimSystem: About to start them going."); }
	
	// Start them all
	SimProcessor p;
	for(i=0; i<nprocs; i++) {
	    p = (SimProcessor)processors.elementAt(i);
	    p.restart();
	}
	
	if (debug) { System.out.println("SimSystem: Running..."); }
	
	// Wait for all to finish
	for(i=0; i<nprocs; i++) { onestopped.p(sk); }
	
	if (debug) { System.out.println("SimSystem: All processors stopped."); }
	
	runStop();
    }

    /** Run an entity for a tick *** NB Not Synchronized! ***
     */
    static public void runEntityTick(SimProcessor p, Entity ent,
				     Time lw, Time rw, Time safe_time) {
        //neosim.kernel.Deadlock_Message ecp3=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp3);                                              
	advanceClock(lw);		// Advancing time.
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp3);
        //neosim.kernel.Deadlock_Message ecp4=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp4);   
	processors.setRunning(p, rw);
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp4);    
        //neosim.kernel.Deadlock_Message ecp6=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp6);           
	EventList el = ent.getEventList();
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp6);
	Time actualT = ent.handleEvents(el, rw);
        
        //neosim.kernel.Deadlock_Message ecp5=new neosim.kernel.Deadlock_Message();
        //neosim.kernel.DeadlockTracker.LockDeadlockTracker(safe_time,ecp5);          
	processors.setStopped(p);
        //neosim.kernel.DeadlockTracker.UnLockDeadlockTracker(ecp5);
    }

}
