package neosim.perf;
 
/** Performance monitoring interface.<p>
 *
 *  This is used for monitoring entity or processor performance,
 *  implementations could include a timing diagram display or a logger
 *  to files.<p>
 *
 *  The basic idea is that a number of types are registered 
 *  (with a name, & a number of states). 
 *  The entity states to monitor are registered
 *  (with a name & typename).
 *  The entities then send state changed events using the
 *  addEvent method.<p>
 *
 *  At present, Strings are used for all states.<p>
 *
 *  Example: monitor state of 2 SimProcessors<p>
 * <pre>
 *    PerfMonitor pm = new TimingDiagram();
 *    int psid = pm.registerType("procstate", "BUSY IDLE");
 *    int p1id = pm.registerEntityState("proc1", "procstate");
 *    int p2id = pm.registerEntityState("proc2", "procstate");
 *    pm.resetEvents();
 *    ... the processors call:
 *        pm.addEvent( p1id, 1.23, "BUSY" );
 *        pm.addEvent( p2id, 3.45, "IDLE" );
 *
 *    pm.finished();
 * </pre> <p>
 *  Typically the ProcManager or SimSystem will register the
 *  types and entity states, and the entities or processors will
 *  call the addEvent() routine.<p>
 * @version	1.0, March 1999
 * @author	Fred Howell
 */

public interface PerfMonitor {
    
    /** Adds a new named type of event, states separated by spaces
     * e.g. <pre>pm.registerType("procstate", "BUSY IDLE WAIT");</pre><p>
     * @param typename The name of the data type.
     * @param types    The values the types can take, separated by spaces,
     *                 e.g. "BUSY IDLE WAIT"
     */
    public int registerType(String typename, String types);

    /** Registers a entity to monitor, with its state type.
     * e.g. <pre>pm.registerEntityState("proc123", "procstate");</pre><p>
     * @param name        The name of the entity.
     * @param typename    The name of its event type, defined above.
     * @return The id of this entity state.
     */
    public int registerEntityState(String name, String typename);

    /** Resets the monitor back to time zero, deleting any events.
     */
    public void resetEvents();

    /** Adds a single event which says that a specified entity 
     * changed state at time t.
     * @param estateId  The id of the entity state
     * @param t  The time of the event
     * @param statename The name of the state, e.g. "BUSY"
     */
    public void addEvent(int estateId, double t, String statename);
    
    /** Resets the monitor, removing any types and entity states.
     */
    public void resetAll();

    /** Tells the monitor that running has finished.
     */
    public void finished();

    /** Tag an extra monitor on the end, all commands get forwarded
     */
    public void daisyChain(PerfMonitor next);
}




 
