package neosim.perf;

import neosim.util.*;
 

import java.util.Vector;

/** Count time intervals in a state, e.g. BUSY 
 */
final class StateStat {
    int count;
    double min;
    double max;
    double total;
    StateStat() { reset(); }
    void reset() { count = 0; min=0; max=0; total=0;}
    double getMin() { return min; }
    double getMax() { return max; }
    double getTotal() { return total; }
    double getCount() { return count; }
    double getAvg() { if (count>0) return total/count; else return 0; }
    /** add an interval */
    void add(double t) { 
	count++; 
	if (count==1) { min=t; max=t; } 
	else if (t<min) { min = t; }
	else if (t>max) { max = t; }
	total += t;
    }
}

/** A vector of State stats for a parameter (BUSY/IDLE/SEND etc) 
 */
final class TypeStats {
    TypeParam tp;
    Vector ts;
    int lastState;
    double lastTime;
    TypeStats( TypeParam tp ) {
	this.tp = tp;
	reset();
    }
    void reset() {
	ts = new Vector();
	for (int i=0; i<tp.getNumLabels(); i++) {
	    ts.addElement( new StateStat() );
	}
	lastState = -1;
	lastTime  = 0.0;
    }

    /** Add a change of state to state at t 
     */
    void addEvent(String state, double t) {
	double interval = t - lastTime;
	int newState = tp.getIndex(state);
	if (lastState!=-1) {
	    StateStat tps = (StateStat)(ts.elementAt(lastState));
	    tps.add(interval);
	}
	lastState = newState;
	lastTime  = t;
    }

    /** Write out 1 line summary per state */
    void report( String title, FileListener fs ) {
	for (int i=0; i<ts.size(); i++) {
	    StateStat tps = (StateStat)(ts.elementAt( i ));
	    String line = title + " " + tp.getLabel(i)+ " " +
		tps.getTotal() + " " +
		tps.getCount() + " " + 
		tps.getMin() + " " + 
		tps.getMax() + " " + 
		tps.getAvg();
	    fs.handleFile( new FileEventObject(this,line) );
	}
    }
}


/** Performance monitoring statistics gatherer.<p>
 *
 *  Implementation of PerfMonitor that writes statistics to a file.
 *  The logging is done as <entity> <time> <state>
 *  The stats are:
 *  &lt;min/max/avg> time in each state
 *
 <pre>
&lt;entitystatename> &lt;total> &lt;count> &lt;min> &lt;max> &lt;avg> in that state
entity1name BUSY 187  100 3 7 5
entity1name IDLE 7 100 0 1 0.1
entity2name BUSY 187 100 4 8 6
entity2name IDLE 4 100 2 1 4 2
</pre>
 */

public final class PerfStats implements PerfMonitor, Filecmds {
    Types  paramTypes;
    Vector states;
    Vector stateStats;  // Vector of TypeStats, 1 per entity state
    FileListener fs;
    String fname;
    double t0;

    PerfMonitor next = null;

    public void daisyChain( PerfMonitor next ) {
	this.next = next;
    }

    public PerfStats(String fname) {
	this.fname = fname;
	resetAll();
    }

    public int registerType(String typename, String values) {
	if (next!=null) { int dummy = next.registerType(typename,values); }
	TypeParam tp = new TypeParam( typename, values );
	paramTypes.add(tp);
	return paramTypes.size()-1;
    }

    public int registerEntityState(String name, String typename) {
	if (next!=null) { int dummy = next.registerEntityState(name,typename); }
	TypeParam tp = paramTypes.find(typename);
	if (tp==null) { 
	    System.out.println("Type "+typename+" not registered");
	    return -1;
	}
	states.addElement( new EntityState(name, tp) );
	stateStats.addElement( new TypeStats(tp) );
	
	return states.size()-1;
    }

    public void fprint(String l) {
	fs.handleFile( new FileEventObject(this,l) );
    }

    /** Write header before events */
    public void resetEvents() {
	if (next!=null) { next.resetEvents(); }
	int i,j;
	String line;
	fs.handleFile( new FileEventObject(this,START) );
	fprint("# Format: <entitystate> <type> <total> <count> <min> <max> <avg>");
    }

    /** A new event 
     */
    public void addEvent(int estateId, double t, String statename) {
	if (next!=null) { next.addEvent(estateId,t,statename); }
	EntityState es = (EntityState)(states.elementAt(estateId));
	TypeStats ts = (TypeStats)(stateStats.elementAt(estateId));
	ts.addEvent(statename,t);
	// Summary over types? not for now.
    }

    public void resetAll() {
	if (next!=null) { next.resetAll(); }
	paramTypes  = new Types();
	states = new Vector(); // of EntityState	
	stateStats = new Vector();
	fs = new FileSaver(fname);
    }

    /** Write stats out to file */
    public void finished() {
	if (next!=null) { next.finished(); }
	/** For each entity state */
	for (int i=0; i<states.size(); i++) {
	    TypeStats ts = (TypeStats)(stateStats.elementAt(i));
	    EntityState es = (EntityState)(states.elementAt(i));
	    ts.report( es.getName(), fs );
	}
	fs.handleFile(new FileEventObject(this,EOF));
    }
}
