package neosim.util;

import neosim.iface.*;
import java.util.Hashtable;

/**
 * Neosim Module Loader - 
 *  utility functions to create Neosim objects from 
 *  jars files.  Calls the Jarloader to read in the jar files.
 * @author Fred Howell
 * @date Nov 2000
 * @updated June 2001 to allow loading from local classpath too
 */
public class ModuleLoader {

    /** Utility functions for extracting parameter values */
    public static int atoi(String s) { return (Integer.valueOf(s)).intValue(); }
    public static double atod(String s) { return (Double.valueOf(s)).doubleValue(); }
    public static int getIntParam( Hashtable ht, String name ) {
	String val = (String)ht.get(name);
	if (val == null) return 0; else return atoi(val);
    }
    public static double getDoubleParam(Hashtable ht, String name) {
	String val = (String)ht.get(name);
	if (val == null) return 0.0; else return atod(val);
    }
    public static boolean getBooleanParam(Hashtable ht, String name) {
	String val = (String)ht.get(name);
	if (val == null) return false; 
	else {
	    if (val.equals("true") || val.equals("1")) { 
		return true;
	    } 	    
	}
	return false;
    }

    
    /** Call JarLoader.loadJarDoOnBean and catch any exceptions
     */
    public static boolean loadJar( String ename, DoOnBean dob ) {
	try { 
	    if (!ename.startsWith("http")&&!ename.startsWith("ftp")){
		if (!ename.startsWith("file:")){
		    String fileename = new String("file:"+ename);
		    JarLoader.loadJarDoOnBean(fileename, dob);
		}
		else JarLoader.loadJarDoOnBean(ename, dob);
	    }
	    else JarLoader.loadJarDoOnBean(ename, dob);
	    return true;
	} catch (Exception e) {
	    System.out.println("Module Loader exception loading"+ename+" "+e);
	    e.printStackTrace();
	    return false;
	    
	}      	
    }

    /** Return an entity class 
     */
    public static EntityClass getEntityClass(String ename, String name, 
					     Hashtable params) {
	EntityClass p = null ;
  	if (ename.startsWith("class:")) {
	    Object o = (Object) makeObject(ename);
	    if (o instanceof neuroml.sim.run.RunnableCellFactory) {
		// make a converter
		p = NeuromlConverter.convertEntityClass((neuroml.sim.run.RunnableCellFactory)o);
	    } else if (o instanceof EntityClass){
		p = (EntityClass)o; 
	    }

	   
	    if (p!=null) 
		p.init(name, params);  
	    return p; 
	} else {

	    EntClassBean eb = new EntClassBean(name, params);
	    loadJar(ename,eb);
	    // ****
	    return eb.getEntityClass();
	}
    }

    /** 
     * Gets an control module from a jar file 
     *
     * @param ename Location of the module
     * @param name Name of the module
     * @param params Hashtable of the parameters for the control module initialization 
     */
    public static ControlModule getControlModule(String ename, Hashtable params) {
	if (ename.startsWith("class:")) {
	    ControlModule p = (ControlModule) makeObject( ename );
	    if (p!=null) 
		p.init(params);  
	    return p; 
	} else {
	    ControlModuleBean cmb = new ControlModuleBean(params);
	    loadJar(ename,cmb);
	    return cmb.getControlModule();
	}
    }

    /** make an object given class: name, e.g.
     * "class:java.lang.String"
     */
    public static Object makeObject( String ename ) {
	Object ret = null;
	String cname = ename.substring(ename.lastIndexOf(":")+1);
	try {
	    Class c = Class.forName( cname );
	    ret = c.newInstance();
	} catch (Exception e) {
	    System.out.println("ModuleLoader readFromClasspath didn't find "+cname+" Error: "+e);	
	}
	return ret;
    }

    /** 
     * Gets a projection jar from a jar file 
     *
     * @param ename Location of the module
     * @param name Name of the module
     * @param params Hashtable of the parameters for the module initialization 
     */
    public static Projection getProjection(String ename, String name,  
		     String srcPop, short outPort, String destPop, short inPort, //PortID outPort//PortID inPort
		     Time delay, Hashtable params) {
	if (ename.startsWith("class:")) {
	    ProjectionJar p = (ProjectionJar) makeObject( ename );
	    if (p!=null) 
		p.init(name, srcPop, outPort, destPop, inPort, delay, params);  
	    return p.getProjection();
	} else {  
	    ProjJarBean pjb = new ProjJarBean(name, srcPop, outPort, destPop, inPort, delay, params);
	    loadJar(ename,pjb);
	    return pjb.getProjectionJar().getProjection();
	}
    }
  /** 
     * Gets a source method jar from a jar file 
     *
     * @param ename Location of the module
     * @param name Name of the module
     * @param params Hashtable of the parameters for the module initialization 
     */
    public static SourceMethod getSourceMethod(String ename, String name,  
		     Hashtable params) {
	if (ename.startsWith("class:")) {
	    SourceMethod p = (SourceMethod) makeObject( ename );
	    if (p!=null) // which params for init ?
		p.init(name, params);  
	    return p;
	} else {  
	    SourceMethodJarBean pjb = new SourceMethodJarBean(name, params);
	    if (!loadJar(ename,pjb))
		return null;
	    else
		return pjb.getSourceMethod();
	}
    }
 /** 
     * Gets a dest method jar from a jar file 
     *
     * @param ename Location of the module
     * @param name Name of the module
     * @param params Hashtable of the parameters for the module initialization 
     */
    public static DestMethod getDestMethod(String ename, String name, Hashtable params) {
	if (ename.startsWith("class:")) {
	    DestMethod p = (DestMethod) makeObject( ename );
	    if (p!=null) // which params for init ?
		p.init(name, params);  
	    return p;
	} else {  
	    DestMethodJarBean pjb = new DestMethodJarBean(name, params);
	    if (!loadJar(ename,pjb))
		return null;
	    else	
		return pjb.getDestMethod();
	}
    }

    /** 
     * Gets a population jar from a jar file 
     *
     * @param ename Location of the module
     * @param name Name of the module
     * @param params Hashtable of the parameters for the module initialization 
     */
    public static Population getPopulation(String ename, String name,  
					   PopulationNode parent, int size, EntityClass ec,
					   EntityInit ei,Hashtable params) {
 	if (ename.startsWith("class:")) {
	    PopulationJar p = (PopulationJar) makeObject( ename );
	    if (p!=null) 
		p.init(name, parent,  size,   ec,   ei, params); 
	    return p.getPopulation();
	} else {  
	    PopJarBean pjb = new PopJarBean(name,  parent,  size,   ec,   ei, params);
	    loadJar(ename,pjb);
	    return pjb.getPopulationJar().getPopulation();
	}
    }
  
}
