package neuroml.core;
import java.util.Hashtable;
import java.util.Enumeration;
/**
 *   derived Unit expressed in terms of the quantities:
 *  M, L, T, charge, temperature, Avagadro's number 
 *  and ten
 *  
 *  Units for intercomponent consistency, mainly for SpatchObject components. 
 * A Unit is defined in terms of powers of the elementary quantities 
 *  mass M (Kg), length L (m), time T (s), charge C (coulombs), 
 *  temperature K (k),  Avagradro's number Navagadro, (~ 6e23) the 
 * electronic charge, elec (coluombs) ten, and a floating coefficient f. 
 * The variables in the Unit object are the powers of these quantities. 
 *  
 *  For example, one volt is one joule per coulomb, which is one newton meter per 
 *   coulomb or one kg m/s _HAT()2/coulomb, so for a Volt Unit, M=1, L=2, s=-2, C=-1
 * and the rest are 0. For mV, one would also set the variable ten to -3.
 * 
 * A micromolarity Unit would have Navagadro=1, L=-3, and 
 * ten = -3 (-6 for micro, +3 for moles per liter instead of per cubic meter).
 * 
 * 
 * The dimensionless quantities Navagadro and ten are not strictly necessary 
 * but are provided for convenience. They make it easier to express quantities
 * of the form Navagadro _HAT()a * 10 _HAT()b .   
 */


public class Unit implements Nameable {
   
    String name; 
    public String getName() { return name; }
    public void setName(String n) { name = n; }    
    
    // dimensions - mass, length, time, coulombs, kelvin
    
    public int M;   
    public int L;
    public int T;
    public int C;
    public int K;
    public int Navagadro;
    public int elec;
    public int ten;
    double f;
    // values min and max determining the possible values range
    double rangeMin;
    double rangeMax;

    
    public Unit () {}
    
    public Unit (String s, int M, int L, int T, int C, int K, 
		 int Nava, int ten, int elec, double f, double min, double max) {
	setName(s);
	this.M = M;
	this.L = L;
	this.T = T;
	this.C = C;
	this.K = K;
	this.Navagadro = Nava;
	this.ten = ten;
	this.elec = elec;
	this.f = f;
	rangeMin = min;
	rangeMax = max;
    }
     
    
    public void setPower (String s, int p) {
	if (s.equals("M")) { M = p;
	} else if (s.equals("L")) { L = p;
	} else if (s.equals("T")) { T = p;
	} else if (s.equals("C")) { C = p;
	} else if (s.equals("K")) { K = p;
	} else if (s.equals("Navagadro")) { Navagadro = p;
	} else if (s.equals("ten")) { ten = p;
	} else {
	    System.out.println ("ERROR - Unit: unknown quantity " + s);
	}
    }  
    
    public static Hashtable<String,Hashtable<String,Unit>> getStandardUnits() {
	Hashtable<String,Hashtable<String,Unit>> u = new Hashtable<String,Hashtable<String,Unit>>();

	// family
	Hashtable<String,Unit> u1 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u2 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u3 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u4 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u5 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u6 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u7 = new Hashtable<String,Unit>();
	Hashtable<String,Unit> u8 = new Hashtable<String,Unit>();
	//                                  M   L   T   C   K  Nav ten  ele   f  min     max
	
	u1.put("Gram",  new Unit("gram",    1,  0,  0,  0,  0,  0, -3,   0,   0., 0.,   1000.));
	u1.put("Kg",    new Unit("Kg",      1,  0,  0,  0,  0,  0,  0,   0,   0., 0.,   1000.));

	u2.put("Second",new Unit("second",  0,  0,  1,  0,  0,  0,  0,   0,   0., 0.,  36000.));
	u2.put("ms",    new Unit("ms",      0,  0,  1,  0,  0,  0, -3,   0,   0., 0.,   1000.));
	
	u3.put("Meter", new Unit("meter",   0,  1,  0,  0,  0,  0,  0,   0,   0., 0.,   1000.));
	u3.put("cm",    new Unit("cm",      0,  1,  0,  0,  0,  0, -2,   0,   0., 0.,    100.));
	u3.put("mm",    new Unit("mm",      0,  1,  0,  0,  0,  0, -3,   0,   0., 0.,    100.));
	u3.put("micron",new Unit("micron",  0,  1,  0,  0,  0,  0, -6,   0,   0., 0.,   1000.));

	u4.put("Volt",  new Unit("Volt",    0,  2, -2, -1,  0,  0,  0,   0,   0.,-1000.,1000.));
	u4.put("mV",    new Unit("mV",      0,  2, -2, -1,  0,  0, -3,   0,   0.,-1000.,1000.));
	
	u5.put("Ampere",new Unit("Ampere",  0,  0, -1,  1,  0,  0,  0,   0,   0., 0.,   1000.));
	u5.put("mA",    new Unit("mA",      0,  0, -1,  1,  0,  0, -3,   0,   0., 0.,   1000.));
	u5.put("nA",    new Unit("nA",      0,  0, -1,  1,  0,  0, -6,   0,   0., 0.,    100.));
	u5.put("pA",    new Unit("pA",      0,  0, -1,  1,  0,  0, -9,   0,   0., 0.,    100.));
	
	u6.put("Joule", new Unit("Joule",   0,  2, -2,  0,  0,  0,  0,   0,   0., 0.,   1000.));
	
	u7.put("Molar", new Unit("Molar",   0, -3,  0,  0,  0,  1,  3,   0,   0., 0.,    100.));
	u7.put("mM",    new Unit("mM",      0, -3,  0,  0,  0,  1,  0,   0,   0., 0.,    100.));
	u7.put("microM",new Unit("microM",  0, -3,  0,  0,  0,  1, -3,   0,   0., 0.,    100.));

	u8.put("none",  new Unit("none",    0,  0,  0,  0,  0,  0,  0,   0,   0., 0.,   1000.));

	u.put("Gram",u1);
	u.put("Second",u2);
	u.put("Meter",u3);
	u.put("Volt",u4);
	u.put("Ampere",u5);
	u.put("Joule",u6);
	u.put("Molar",u7);
	u.put("None",u8);
	return u;
    }  
    // create an unit by name
    public static Unit getStandardUnit(String name) {
	Hashtable<String,Hashtable<String,Unit>> units = getStandardUnits();
	for (Enumeration enum_ = units.keys() ; enum_.hasMoreElements() ;) {
	    String Key = (String)enum_.nextElement();
	    Hashtable<String,Unit> u = units.get(Key);
	    if(u.containsKey(name))
		return u.get(name);
	}
	// else
	return (units.get("None")).get("none");
    }


    public static Hashtable<String,Unit> getStandardUnits(String family) {
	Hashtable<String,Hashtable<String,Unit>> units = getStandardUnits();
	return units.get(family);

    }
    

    public static String getUnitFamily(String name){
	Hashtable<String,Hashtable<String,Unit>> units = getStandardUnits();
	for (Enumeration enum_ = units.keys() ; enum_.hasMoreElements() ;) {
	    String Key = (String)enum_.nextElement();
	    Hashtable<String,Unit> u = units.get(Key);
	    if(u.containsKey(name))
		return Key;
	}
	// else
	return new String("None");
    }

    public double getMinValue(){
	return rangeMin;
    }
    public double getMaxValue(){
	return rangeMax;
    }
}
