package neuroml.sim.run;
 
import neuroml.sim.state.*;

/** Interface for a pre-synaptic terminal.
 * Modules which implement models of pre synaptic
 * vesicle release can implement this interface.
 *
 * example main cell code (where the spikes arrive...)
 *<pre>

class myRunnableCell implements RunnableCell {

void handleSpike( t, cid ) {
  Object o = presyn[cid].handleSpike(t);
  if (o !=null) {
    // Pass message from presyn to postsyn terminal
    // cleft delay dt
    postSyn[cid].handleNeurotransmitter(t+dt, o, cid);
  }
}

// Example cell advance logic
void advance( t ) {
  for (tt = currentTime; tt < t; tt += timestep) {
    for (i=0; i<numSyns; i++) {
      // update internal state of pre/post synapses
      presyn[i].advance(tt);
      postsyn[i].advance(tt);
    }

    // for j=0; j<numConpartments; j++
    //   compartment[j].advance(timestep)

    // Check for spike threshold.
    // if Vmemb > 0 sendSpike()
    //    backpropogate spike to postsynapses ?
    //    maybe send retrograde messages to presynapses ?
  }
}

}
 *</pre>
 */
public interface PreSynapse {

    /** Handle a spike. 
     * This will be called by the system when the spike arrives at the end of
     * the axon, just before the pre synaptice terminal.
     * It can trigger internal changes to the presynaptic bouton.
     * @param t the spike arrival time
     * @return a value indicating (e.g.) amount of neuroxmitter release, or null if
     *            this spike does not result in release.
     */
    public Object handleSpike(double t);

    /** Handle retrograde message. 
     * This can be called by the post synaptic model to send a signal
     * back to this presynapse.
     *
     * @param t the retrograde message arrival time
     * @param o the retrograde message 
     */
    public void handleRetrogradeMessage(double t, Object o);

    /** Advance internal state of this pre synaptic terminal
     */
    public void advance(double t);

    /** Return state of this synapse.
     */
    public SynapseState getState();
}
