package neuroml.util;

import neuroml.core.*;

import java.lang.reflect.*;
import java.util.Vector;
import java.util.Collection;
import java.util.Map;
import java.util.Iterator;
import java.io.*;

/** Write out XML given an object - SBML style
 */
abstract public class XMLOut {

    static int pageWidth = 100; // For prettyprinting word wrap

    /** A Field to choose whether to use attributes or elements for base types
     */
    static boolean useAttributes = true;
    
    /** A Field to choose whether to use field or type name in set elements
     * <pre>
     * false:- Set things = new Set("Widget"); -->
     * &lt;things>
     *    &lt;thing />
     *    &lt;thing class="MyWidget"/>
     * &lt;/things>
     *
     * true:-
     * &lt;things>
     *    &lt;Widget/>
     *    &lt;MyWidget/>
     * &lt;/things>
     * </pre>
     */
    static boolean useTypeNameInSet = false;

    /** A Field to choose whether to use field or type name in set elements
     * <pre>
     * false:- Set things = new Set("Widget"); -->
     *    &lt;thing />
     *    &lt;thing class="MyWidget"/>
     *
     * true:-
     * &lt;things>
     *    &lt;Widget/>
     *    &lt;MyWidget/>
     * &lt;/things>
     * </pre>
     */
    static boolean useSetName = true;

    /** Vector of known package names. Added to search path.
     * e.g. "neuroml"
     */
    static Vector<String> knownPackages = new Vector<String>();
    
    /** A top level URL for the XML file 
     * Will translate into <pre>"&lt;neuroml class="a.b.c" nmlschema="file:./moduleCache/myclasses.jar"/></pre>
     */
    static String nmlschema = null;
    
    /** Set the top level schema location (url of a .jar file)
     * @param s if set, the nmlschema= attribute will be added.
     */
    public static void setNmlSchema(String s) { nmlschema = s; }

    /** Set whether to use attributes or elements for base types (int/double/string)
     * @param f if set true, then base types will be written in attributes
     */
    public static void setUseAttributes(boolean f) { useAttributes = f; }

    /** Set whether to write out the Set name or not (default true)
     * @param f if set false, then Set name will not be written
     */
    public static void setUseSetName(boolean f) { useSetName = f; }


    /** Add a package to the recognised list
     * This is used to allow short class names,
     * e.g. instead of class="org.mysimulator.MyClass"
     * it can write "class=MyClass"
     *<p>
     * e.g. addKnownPackage("org.mysimulator");
     * @param pname package name, e.g. "org.mysimulator"
     */
    public static void addKnownPackage(String pname) { knownPackages.addElement(pname); }

    /** Set whether to use the class name as the &lt;element> in a set
     * @param f if set true, then set members written as &lt;Type> rather
     *    than &lt;setfieldname {class="Type"}>
     */
    public static void setUseTypeNameInSet(boolean f) { useTypeNameInSet = f; }

    /** Write all fields 
     */
    public static String toString(Object top) {
	StringBuffer sb = new StringBuffer();
	String toptag,bottag;
	if (useTypeNameInSet) {
	    toptag = "<"+chopClass(top.getClass().getName());
	    bottag = "</"+chopClass(top.getClass().getName()+">");
	} else {
	    toptag = "<neuroml class=\""+chopClass(top.getClass().getName())+"\"";
	    bottag = "</neuroml>";
	}
	if (nmlschema!=null) {
	    if (!"".equals(nmlschema)) {
		toptag=toptag+" nmlschema=\""+nmlschema+"\"";
	    }
	}
	writeObject(top, sb, top.getClass().getFields(), 0,toptag, bottag,null,null,null);
	return sb.toString();
    }
    
    /** Write all fields to a file 
     */
    public static void toFile(Object top, String filename){
        try {
            FileWriter fw = new FileWriter(filename);
            fw.write(toString( top ));
            fw.close();
        } catch (IOException e) {
            System.err.println("Unable to open file " + e.getMessage());
        }
    }

    /** Write all fields, user specified top level tags
     */
    public static String toString(Object top, String topTag, String bottomTag) {
	StringBuffer sb = new StringBuffer();
	writeObject(top, sb, top.getClass().getFields(), 0, topTag, bottomTag, null,null,null);
	return sb.toString();
    }

    /** Write selected fields
     */
    public static String toString(Field[] selectedFields, Object top) {
       StringBuffer sb = new StringBuffer();
       writeObject(top, sb, selectedFields, 0);
       return sb.toString();
    }

      
    static String indentString(int i) {
	char[] chars = new char[i*3];
	for (int j=0; j<chars.length; j++) { chars[j]=' '; }
	return new String(chars);
    }
    
    static void appendQuoted(StringBuffer sb, String s) { 
       sb.append("\""); 
       if (s!=null) sb.append(s); 
       sb.append("\""); 
    }

    static void appendNVPair(int indentLevel, StringBuffer sb, String type, String name, String value) { 
       sb.append(indentString(indentLevel + 1));
       sb.append("<"+name+">");
       sb.append(value);
       sb.append("</"+name+">\n");
    }

    /** Append name="value" attr 
     */
    static void reallyAppendNVAttr(int indentLevel, StringBuffer sb, String type, String name, String value) { 
 	int newlinelen = sb.length() - sb.toString().lastIndexOf("\n") +4;
	if (name==null || value==null) return;
	newlinelen += name.length()+ value.length();
	if (newlinelen > pageWidth) { 
	    sb.append("\n"+indentString(indentLevel + 2));
	} else { 
	    sb.append(" "); 
	}
	sb.append(name+"=");
	appendQuoted(sb,value);   
    }

    
    /** Append name="value" attr  unless useAttributes = false
     * Version for base elements of a list.
     */
    static void elemAppendNVAttr(int indentLevel, StringBuffer sb, String type, String name, String value) { 
	if ((value!=null) && !"".equals(value) ) {
	    if (useAttributes) { reallyAppendNVAttr( indentLevel, sb, type, name, value ); }
	    else {
		sb.append(value); // The tags will be done by writeObject...
	    }
	}
    }

    /** Append name="value" attr  unless useAttributes = false
     */
    static void appendNVAttr(int indentLevel, StringBuffer sb, String type, String name, String value) { 
	if ((value!=null) && !"".equals(value) ) {
	    if (useAttributes) {		
		reallyAppendNVAttr( indentLevel, sb, type, name, value );
	    } else { // Write attributes as elements
		appendNVPair( indentLevel, sb, type, name, value );
	    }
	}
    }
    
    /** Append name="valueunit" attr 
     */
    static void appendNVUAttr(int indentLevel, StringBuffer sb, String type, 
			      String name, String value, String unit) { 	
	appendNVAttr(indentLevel, sb, type, name, value+unit);
    }
     
    static void appendNVU(int indentLevel, StringBuffer sb, String type, 
			 String name, String value, String unit) { 
       sb.append(indentString(indentLevel + 1));
       sb.append("<"+name+">");
       sb.append(value);
       sb.append(unit);
       sb.append("</"+name+">\n");
    }     

    static void writeObject (Object ob, 
			     StringBuffer sb, 
			     int indentLevel) {
	writeObject( ob, sb, ob.getClass().getFields(), indentLevel,null,null,null,null,null);
    }
    static void writeObject (Object ob, 
			     StringBuffer sb, 
			     int indentLevel,
			     String defClass,
			     String nameOfSet,
			     String nameOfField) {
	writeObject( ob, sb, ob.getClass().getFields(), indentLevel,null,null,defClass,nameOfSet,nameOfField);
    }

    static void writeObject (Object ob, 
			     StringBuffer sb, 
			     Field[] selectedFields, 
			     int indentLevel) {
	writeObject( ob, sb, selectedFields, indentLevel,null,null,null,null,null);
    }
    
    /** Return a class name 
     * If it starts with "neuroml." or "java.lang." it returns
     * just the short class name, otherwise it returns the full path.
     */
    public static String chopClass(String cname) {
	if (cname == null) { return ""; }
	if (cname.startsWith("neuroml.") || cname.startsWith("java.lang.")) {
	    cname = cname.substring(cname.lastIndexOf(".")+1);
	} else {
	    boolean found = false;
	    // Slow algorithm for chopping class names given knownPackages...
	    for (int i=0; (i<knownPackages.size()) && !found; i++) {
		if ( cname.startsWith((knownPackages.elementAt(i))+".") ) {
		    cname = cname.substring(cname.lastIndexOf(".")+1);
		    found = true;
		}
	    }
	}
	return cname;
    }
    
    /** Capitalize 1st character of a string */
    public static String capitalize(String s) {
	return s.substring(0,1).toUpperCase() + s.substring(1,s.length());
    }

    /** Make plural singular - by removing trailing "s" from a string 
     * (not very international yet)
     */
    public static String makeSingular(String s) {	
	if (s.endsWith("cies")) return s.substring(0,s.length()-1);
	else if (s.endsWith("ies")) return s.substring(0,s.length()-3)+"y";
	else if (s.endsWith("sses")) return s.substring(0,s.length()-2);
	else if (s.endsWith("s")) return s.substring(0,s.length()-1);
	else return s;
    }
    
    /**
     * @param selectedFields fields to write out
     */
    static void writeObject( Object ob, 
			     StringBuffer sb, 
			     Field[] selectedFields, 
			     int indentLevel,
			     String topTag, // optional replacement tags
			     String bottomTag,
			     String defaultClassName,  // if object not of this class, add class= tag
			     String nameOfSet,
			     String nameOfField       
			     ) {
	String setPrefix  = ""; // ... or "setOf"
	String listPrefix = ""; // ... or "listOf"
	boolean docaps = false; // setOfdimensions or setOfDimensions
	StringBuffer contents = new StringBuffer(); // For contents of xml
	StringBuffer header   = new StringBuffer(); // For attr list
	StringBuffer tailer   = new StringBuffer(); // For attr list
	
	header.append(indentString(indentLevel));
	tailer.append(indentString(indentLevel));
	String cname = chopClass( ob.getClass().getName() );
	String objtag="void";
	boolean classSpecified = false;

	if (topTag==null) {
	    header.append("<");	  

	    // Is it member of a set or list?
	    if (nameOfSet!=null) {
		if (useTypeNameInSet) {
		    objtag = cname;
		    header.append(cname); // Just use short class name.... 
		} else {
		    objtag=makeSingular(nameOfSet);
		    header.append(objtag);
		}
		
		// This is the wrong way to do named sets / hashtables....
		// fwh commented this out as a test Apr 2002
		//if (ob instanceof Nameable) {
		//    reallyAppendNVAttr(indentLevel,header, "","name",((Nameable)ob).getName());
		//}
	    } else {
		if (nameOfField!=null) { 
		    objtag = nameOfField;
		} else {
		    if (ob instanceof Nameable) {
			objtag = ((Nameable)ob).getName();
		    } else { // If not standard class
			objtag="object";
		    }
		}
		header.append(objtag);
	    }	    

	    // Is class derived?
	    //	    System.out.println("cname "+cname+" defaultClassName "+defaultClassName+" chopped "+
	    // chopClass(defaultClassName));
	    // This class comparison isn't perfect - it only checks the class name, not the full path	    
	    if (defaultClassName==null || !(cname.equals(chopClass(defaultClassName)))) {
		if ((nameOfSet!=null) && useTypeNameInSet) {
		    // Don't put the class name here!!! as it's already in the <element>
		} else {
		    reallyAppendNVAttr(indentLevel, header,"","class",cname);
		    classSpecified = true;
		}
	    }	    
	} else {
	    header.append(topTag);
	}
	
	// If not using attributes, the header is now done...
	if (!useAttributes) { header.append(">"); }

	boolean onlyBasic = false;// true if this object is a basic one...
	// Check if ob is a basic type (can happen if ob is from a list or a set)
	String attribs = "value";
	if (ob instanceof Integer ) {
	    elemAppendNVAttr( indentLevel, header, "int", attribs, ob.toString() );
	    onlyBasic = true;
	} else if (ob instanceof Short ) {
	    elemAppendNVAttr( indentLevel, header, "short", attribs, ob.toString() );
	    onlyBasic = true;
	} else if (ob instanceof Double) {
	    elemAppendNVAttr( indentLevel, header, "double", attribs, ob.toString() );
	    onlyBasic = true;
	} else if (ob instanceof Float) {
	    elemAppendNVAttr( indentLevel, header, "float", attribs, ob.toString() );
	    onlyBasic = true;
	} else if (ob instanceof Boolean) {
	    elemAppendNVAttr( indentLevel, header, "bool", attribs, ob.toString() );
	    onlyBasic = true;
	} else if (ob instanceof neuroml.core.Enum) {
	    elemAppendNVAttr( indentLevel, header, "Enum", attribs, ob.toString() );
	    onlyBasic = true;
	} else if (ob instanceof Reference) {
	    elemAppendNVAttr( indentLevel, header, "Reference", attribs, 
			      ((Reference)ob).getTargetName() );
	    onlyBasic = true;
	} else if (ob instanceof String) { // *** Bad news if set of string... ***
	    String s = (String)ob;
	    // If any funny characters, put in contents, else in attr.
	    if ((s.indexOf("\n")>0) || s.indexOf("\"")>0) {
		contents.append(indentString(indentLevel + 1));
		contents.append("<"+objtag+">");
		if (s.indexOf("\n")>0) { contents.append("\n"); }
		contents.append(Extend.toQuoted(s));
		if (s.indexOf("\n")>0) { contents.append(indentString(indentLevel + 1)); }
		contents.append("</"+objtag+">\n"); 
	    } else {
		appendNVAttr(indentLevel, header, "String", attribs, ob.toString() );
	    }
	} else if (ob instanceof java.util.Collection) {
	    // Whoops, printing out a top level collection...
	    // This only happens if this is in a set of sets
	    // See field.write below
	    Collection set = (Collection)ob;
	    int subtab = 2;
	    // Convert set name...
	    String sn = nameOfSet; // ??
	    String stag = setPrefix+(docaps ?capitalize(sn) : sn);
	    subtab = 1; 
	    String elType = "java.lang.Object";
	    if (set instanceof ContentRestricted) { elType = ((ContentRestricted)set).getContentClassName(); }
	    Iterator it = set.iterator();
	    while (it.hasNext()) {
		Object sub = it.next();
		writeObject (sub, contents, indentLevel+subtab, elType, sn, null);
	    }	
	} else {  
	    if (!useAttributes) { header.append("\n"); }
	    if (ob instanceof Nameable) {
		appendNVAttr( indentLevel, header, "int", "name", ((NamedObject)ob).getName() );
	    }
	    Field[] flds = selectedFields;
	    for (int i = 0; i < flds.length; i++) {
		Object obj = null;
		int modif = flds[i].getModifiers();
		
		if (Modifier.isPublic( modif ) && !Modifier.isStatic( modif )) {
		    try {
			obj = flds[i].get(ob);
		    } catch (Exception e) {
			System.out.println("XMLOut: reflection error");
			e.printStackTrace();
			return;
		    }		    
		    if (obj == null) {
			// do nothing
		    } else  if (obj instanceof Integer) {
			appendNVAttr( indentLevel, header, "int", flds[i].getName(), obj.toString() );
		    } else  if (obj instanceof Short) {
			appendNVAttr( indentLevel, header, "short", flds[i].getName(), obj.toString() );
		    } else  if (obj instanceof Double) {
			appendNVAttr( indentLevel, header, "double", flds[i].getName(), obj.toString() );
		    } else  if (obj instanceof Float) {
			appendNVAttr( indentLevel, header, "float", flds[i].getName(), obj.toString() );
		    } else if (obj instanceof DValue) {
			DValue n = (DValue) obj;
			appendNVUAttr( indentLevel, header, "DValue", flds[i].getName(), 
				       ""+(n.getValue()),n.getUnitRef().getTargetName()  );			
		    } else if (obj instanceof Boolean) {
			appendNVAttr( indentLevel, header, "bool", flds[i].getName(), obj.toString() );
		    } else if (obj instanceof neuroml.core.Enum) {
			appendNVAttr( indentLevel, header, "Enum", flds[i].getName(), obj.toString() );
		    } else if (obj instanceof String) {
			String s = (String)obj;
			// If any funny characters, put in contents, else in attr.
			if ((s.indexOf("\n")>0) || s.indexOf("\"")>0) {
			    contents.append(indentString(indentLevel + 1));
			    contents.append("<"+flds[i].getName()+">");
			    if (s.indexOf("\n")>0) { contents.append("\n"); }
			    contents.append(Extend.toQuoted(s));
			    if (s.indexOf("\n")>0) { contents.append(indentString(indentLevel + 1)); }
			    contents.append("</"+flds[i].getName()+">\n"); 
			} else {
			    appendNVAttr( indentLevel, header, "String", flds[i].getName(), obj.toString() );
			}
		    } else if (obj instanceof DoubleArray) {
			DoubleArray da = (DoubleArray)obj;
			double[] d =  da.getData();
			if (d != null) {
			    contents.append(indentString(indentLevel + 1)+"<"+flds[i].getName());
			    da.setName(flds[i].getName());
			    String u = da.getUnitRef().getTargetName();
			    if (!"".equals(u)) contents.append(" unit=\""+da.getUnitRef().getTargetName()+"\"");
			    contents.append(" length=");
			    appendQuoted(contents,""+d.length ); 
			    contents.append(">\n");
			    /*
			      <state unit="us" length="10">
			      1.34 232 44 54 ...
			      </state>
			    */
			    for (int j=0; j<d.length; j++) {
				contents.append(""+d[j]+"\t");
				if ((j+1)%10==0 && (j+1!=d.length) ) {contents.append("\n"); }
			    }
			    contents.append("\n"+indentString(indentLevel + 1)+"</"+flds[i].getName()+">\n");
			}
		    } else if (obj instanceof IntArray) { 
			IntArray da = (IntArray)obj;
			int[] d =  da.getData();
			contents.append(indentString(indentLevel + 1)+"<"+da.getName());
			da.setName(flds[i].getName());
			contents.append(" length=");
			appendQuoted(contents,""+d.length ); 
			contents.append(">\n");
			
			for (int j=0; j<d.length; j++) {
			    contents.append(""+d[j]+"\t");
			    if ((j+1)%10==0) {contents.append("\n"); }
			}
			contents.append("\n"+indentString(indentLevel + 1)+"</"+da.getName()+">\n");
		    } else if (obj instanceof StringArray) {
			StringArray da = (StringArray)obj;
			String[] d =  da.getData();
			da.setName(flds[i].getName());
			contents.append(indentString(indentLevel + 1)+"<"+da.getName());
			contents.append(" length=");
			appendQuoted(contents,""+d.length ); 
			contents.append(">\n");
			
			for (int j=0; j<d.length; j++) {
			    contents.append(""+d[j]+"\t");
			    if ((j+1)%10==0) {contents.append("\n"); }
			}
			contents.append("\n"+indentString(indentLevel + 1)+"</"+da.getName()+">\n");
                    } else if (obj instanceof java.util.Collection) {
                        
			Collection set = (Collection)obj;
			String sn = flds[i].getName();
			if (set instanceof Nameable) { ((Nameable)set).setName(sn); }

			int subtab = 2;
			// Convert set name...
			String stag = setPrefix+(docaps ?capitalize(sn) : sn);
                        
                        
                        
                         

			if (useSetName) {
			    contents.append(indentString(indentLevel + 1));
			    contents.append("<"+stag);
                            //JAG
                            Field[] flds2 = obj.getClass().getFields();
                            for (int l = 0; l < flds2.length; l++) {
                                Object obj2 = null;
                                int modif2 = flds2[l].getModifiers();
		
                                if (Modifier.isPublic( modif2 ) && !Modifier.isStatic( modif2 )) {
                                    try {
                                        obj2 = flds2[l].get(obj);
                                    } catch (Exception e) {
                                        System.out.println("XMLOut: reflection error");
                                        e.printStackTrace();
                                        return;
                                    }		    
                                    if (obj2 == null) {
			// do nothing
                                    } else  if (obj2 instanceof Integer ) {
                                        appendNVAttr( indentLevel, contents, "int", flds2[l].getName(), obj2.toString() );
                                    } else  if (obj2 instanceof Short ) {
                                        appendNVAttr( indentLevel, contents, "short", flds2[l].getName(), obj2.toString() );
                                    } else  if (obj2 instanceof Double) {
                                        appendNVAttr( indentLevel, contents, "double", flds2[l].getName(), obj2.toString() );
                                    } else  if (obj2 instanceof Float) {
                                        appendNVAttr( indentLevel, contents, "float", flds2[l].getName(), obj2.toString() );
                                    } else if (obj2 instanceof DValue) {
                                        DValue n = (DValue) obj2;
                                        appendNVUAttr( indentLevel, contents, "DValue", flds2[l].getName(), 
                                            ""+(n.getValue()),n.getUnitRef().getTargetName()  );			
                                    } else if (obj2 instanceof Boolean) {
                                        appendNVAttr( indentLevel, contents, "bool", flds2[l].getName(), obj2.toString() );
                                    } else if (obj instanceof neuroml.core.Enum) {
                                        appendNVAttr( indentLevel, contents, "Enum", flds2[l].getName(), obj2.toString() );
                                    } else if (obj2 instanceof String) {
                                        String s = (String)obj;
			// If any funny characters, put in contents, else in attr.
                                        if ((s.indexOf("\n")>0) || s.indexOf("\"")>0) {
                                            contents.append(indentString(indentLevel + 1));
                                            contents.append("<"+flds2[l].getName()+">");
                                            if (s.indexOf("\n")>0) { contents.append("\n"); }
                                            contents.append(Extend.toQuoted(s));
                                            if (s.indexOf("\n")>0) { contents.append(indentString(indentLevel + 1)); }
                                            contents.append("</"+flds2[l].getName()+">\n"); 
                                        } else {
                                            appendNVAttr( indentLevel, header, "String", flds2[l].getName(), obj2.toString() );
                                        }
                                    }
                                }
                            }                        
                            contents.append(">\n");
			} else { subtab = 1; }
			String elType = "java.lang.Object";
			if (set instanceof ContentRestricted) { elType = ((ContentRestricted)set).getContentClassName(); }
			//			String elType = set.getType();
			
			Iterator it = set.iterator();
			while (it.hasNext()) {
			    Object sub = it.next();
			    //			    System.out.println("Doing writeobject of "+sn);
			    writeObject(sub, contents, indentLevel+subtab, elType, sn, null);
			}		
			if (useSetName) {
			    contents.append(indentString(indentLevel + 1));
			    contents.append("</"+stag+">\n");
			}			
		    } else if (obj instanceof java.util.Map) {
			Map map = (Map)obj;
			String sn = flds[i].getName();
			if (map instanceof Nameable) { ((Nameable)map).setName(sn); }

			int subtab = 2;
			// Convert set name...
			String stag = setPrefix+(docaps ?capitalize(sn) : sn);
			if (useSetName) {
			    contents.append(indentString(indentLevel + 1));
			    contents.append("<"+stag+">\n");
			} else { subtab = 1; }
			String elType = "java.lang.Object";
			if (map instanceof ContentRestricted) { elType = ((ContentRestricted)map).getContentClassName(); }
			//			String elType = set.getType();
			
			java.util.Set keys = map.keySet();
			Iterator it = keys.iterator();
			while (it.hasNext()) {
			    Object k = it.next();
			    Object sub = map.get(k);
			    // Use name=k.toString()...
			    String kn = sn+" name=\""+k+"\"";
			    if (! sub.getClass().getName().equals( elType )) {
				kn = kn+" class=\""+chopClass(sub.getClass().getName())+"\"";
			    }
			    writeObject(sub, contents, sub.getClass().getFields(), 
					indentLevel+subtab,"<"+kn,"</"+sn+">",elType,sn,null);
			    //			    writeObject (sub, contents, indentLevel+subtab, elType, kn, null);
			}		
			if (useSetName) {
			    contents.append(indentString(indentLevel + 1));
			    contents.append("</"+stag+">\n");
			}
		    } else if (obj instanceof Reference) {
			Reference ref = (Reference)obj;
			if (ref instanceof Nameable) { ((Nameable)ref).setName(flds[i].getName()); }
			appendNVAttr( indentLevel, header, "double", flds[i].getName(), ref.getTargetName() ); 	
		    } else if (obj instanceof Object) {
                    //if (obj instanceof Nameable) {
                    //    System.out.println("setting name to "+flds[i].getName());
                    //    ((Nameable)obj).setName(flds[i].getName());
                    //}
			String expectedClass = flds[i].getType().getName();
			writeObject((Object)obj, contents, indentLevel+1, expectedClass, null, flds[i].getName()); 
		    } else {
			System.out.println("ERROR - XMLOut: unknown object type " + obj +" "+flds[i].getName() );
		    }
		}
	    }
	} //end if ob is not a basic type
	
	if (bottomTag==null) {
	    tailer.append("</"+objtag+">\n");	  
	} else {
	    tailer.append(bottomTag+"\n");
	}

	// Only try shorthand if using attributes
	if (useAttributes) {
	    if (contents.length()==0) {
		header.append("/>\n"); // if empty, use shorthand
		tailer= new StringBuffer();
	    } else {
		header.append(">\n");
	    }
	    sb.append(header);
	    sb.append(contents);
	    sb.append(tailer);
	} else {
	    if (onlyBasic) { // Special case for basic objects in a set/list
		sb.append(header);sb.append("</"+objtag+">\n");
	    } else {
		sb.append(header);
		sb.append(contents);
		sb.append(tailer);
	    }
	}
    }	    
}






